package teddy;

import java.awt.*;
import java.util.*;
import java.math.*;



public class Delaunay {
	
	public static int[][] triangulate_for_Plate(Vector loop){
		Vector _loop = new Vector();
		for(int i=0; i<loop.size(); i++){
			double[] xy = (double[]) loop.elementAt(i);
			_loop.addElement(new Vector2(xy[0], xy[1]));
		}
		return (new Delaunay()).triangulate_main(_loop);
	}
	
	Delaunay(){};
	
	Vector triangles;
	Vector edges;
	public int[][] triangulate_main(Vector loop){
		triangles = new Vector();
		edges = new Vector();
		Vertex v0 = new Vertex((Vector2)loop.elementAt(0),0);
		Vertex v1 = new Vertex((Vector2)loop.elementAt(1),1);
		Vertex v2 = new Vertex((Vector2)loop.elementAt(2),2);
		Edge e0 = add(new Edge(v0, v1, true));
		Edge e1 = add(new Edge(v1, v2, true));
		Edge e2 = add(new Edge(v2, v0, false));
		Triangle t0 = add(new Triangle(e0,e1,e2));
		
		baseVertex = v0;
		prevVertex = v2;
		prevEdge = e2;
		for(int i=3; i<loop.size(); i++){
			Vertex v = new Vertex((Vector2)loop.elementAt(i), i);
			add_vertex(v);
		}
		
		/*
		Vector loops = new Vector();
		for(int i=0; i<triangles.size(); i++){
			Triangle t = (Triangle) triangles.elementAt(i);
			loop = new Vector();
			loop.addElement(t.get_vertex(0).vector2());
			loop.addElement(t.get_vertex(1).vector2());
			loop.addElement(t.get_vertex(2).vector2());
			loops.addElement(loop);
		}
		return loops;	
		*/
		int[][] result = new int[triangles.size()][3];
		for(int i=0; i<triangles.size(); i++){
			Triangle t = (Triangle) triangles.elementAt(i);
			result[i][0] = t.get_vertex(0).index;
			result[i][1] = t.get_vertex(1).index;
			result[i][2] = t.get_vertex(2).index;
		}
		return result;	
		
	}
	public Vertex baseVertex;
	public Vertex prevVertex;
	public Edge prevEdge;
	public void add_vertex(Vertex v){
		Edge edge0 = add(new Edge(baseVertex, v, false));
		Edge edge1 = add(new Edge(prevVertex, v, true));
		Triangle triangle = add(new Triangle(prevEdge, edge1, edge0));
		
		latest_vertex = v;
		clearFlipped();
		recursiveFlip(prevEdge, null);
		
		clearFlipped();
		for(int i=0; i<edges.size(); i++){
			Edge edge = (Edge)edges.elementAt(i);
			if (double_flipped(edge)){
				recursiveFlip(edge, null);
				clearFlipped();
			}
		}
		
		prevVertex = v;
		prevEdge = edge0;
	}
	public void clearFlipped(){
		for(int i=0; i<edges.size(); i++)
			((Edge)edges.elementAt(i)).flipped = false;
	}
	public void recursiveFlip(Edge edge, Edge fixed){
		if (edge.constrained)
			return;
			
		Triangle triangle0 = edge.left;
		Triangle triangle1 = edge.right;
		
		if (triangle0 == null || triangle1 == null)
			return;
			
		Vertex v0 = triangle0.get_the_other_vertex(edge);
		Vertex v1 = triangle1.get_the_other_vertex(edge);
		
		if (should_flip(edge, v0, v1)){
			Edge edge0s = triangle0.get_the_other_edge(edge.end);
			Edge edge0e = triangle0.get_the_other_edge(edge.start);
			Edge edge1s = triangle1.get_the_other_edge(edge.end);
			Edge edge1e = triangle1.get_the_other_edge(edge.start);
			
			delete(triangle0);
			delete(triangle1);
			delete(edge);


			Edge flippeEdge = add(new Edge(v0, v1, false));
			flippeEdge.flipped = true;
			triangle0 = add(new Triangle(flippeEdge, edge1e, edge0e));
			triangle1 = add(new Triangle(flippeEdge, edge0s, edge1s));
			
			if (edge0s != fixed)
				recursiveFlip(edge0s, edge);
			if (edge0e != fixed)
				recursiveFlip(edge0e, edge);
			if (edge1s != fixed)
				recursiveFlip(edge1s, edge);
			if (edge1e != fixed)
				recursiveFlip(edge1e, edge);
		}
	}
	public double get_angle(Vertex v0, Vertex v1, Vertex v2){
		Vector2 vec0 = new Vector2(v1, v0);
		Vector2 vec1 = new Vector2(v1, v2);
		return vec0.get_relative_angle(vec1);
	}
	public double min(double a, double b, double c, double d){
		return Math.min(Math.min(a,b),Math.min(c,d));
	}
	public boolean same_side(Vertex start, Vertex end, Vertex v0, Vertex v1){
		Vector2 axis = new Vector2(start, end);
		Vector2 vec0 = new Vector2(start, v0);
		Vector2 vec1 = new Vector2(start, v1);
		return (axis.cross_product(vec0) * axis.cross_product(vec1) > 0);
	}
	Vertex latest_vertex;
	public boolean double_flipped(Edge edge){
		if (edge.flipped)
			return false;
		if (edge.constrained)
			return false;
		Triangle triangle0 = edge.left;
		Triangle triangle1 = edge.right;
		if (triangle0 == null || triangle1 == null)
			return false;
		Vertex v0 = triangle0.get_the_other_vertex(edge);
		Vertex v1 = triangle1.get_the_other_vertex(edge);
		return (same_side(edge.start, edge.end, v0, v1));// &&same_side(v0, v1, edge.start, edge.end));
	}
	public boolean should_flip(Edge edge, Vertex v0, Vertex v1){
		if (edge.flipped) return false;
		
		if (same_side(edge.start, edge.end, v0, v1)){
			return true;
		}
		
		if (same_side(v0, v1, edge.start, edge.end))
			return false;
			
		double originalMinimumAngle = min(
										get_angle(edge.start, edge.end, v0),
										get_angle(edge.end, edge.start, v0),
										get_angle(edge.start, edge.end, v1),
										get_angle(edge.end, edge.start, v1));
		double flippedMinimumAngle = min(
										get_angle(edge.start, v0, v1),
										get_angle(edge.start, v1, v0),
										get_angle(edge.end, v0, v1),
										get_angle(edge.end, v1, v0));
		
		return (flippedMinimumAngle > originalMinimumAngle);
	}	
	
	public Triangle add(Triangle triangle){
		triangles.addElement(triangle);
		return triangle;
	}
	public Edge add(Edge edge){
		edges.addElement(edge);
		return edge;
	}
	public void delete(Triangle triangle){
		triangles.removeElement(triangle);
		triangle.delete();
	}
	public void delete(Edge edge){
		edges.removeElement(edge);
		edge.delete();
	}
	
	
	
	class Vertex extends Vector2{
		int index;
		Vector edges = new Vector();
		Vertex(Vector2 v, int n){
			x = v.x;
			y = v.y;
			index = n;
		}
		public void remove(Edge edge){
			edges.removeElement(edge);
		}
		public Vector2 vector2(){
			return new Vector2(x,y);
		}
	}
	class Edge {
		Triangle left;
		Triangle right;
		Vertex start;
		Vertex end;
		boolean constrained;
		boolean flipped;
		Edge(Vertex v0, Vertex v1, boolean c){
			start = v0;
			end = v1;
			constrained = c;
			
			start.edges.addElement(this);
			end.edges.addElement(this);
		}
		public void delete(){
			start.remove(this);
			end.remove(this);
		}
		public void set_left(Triangle triangle){
			if (left != null)	System.out.println("Error in Edge.set_left");
			left = triangle;
		}
		public void set_right(Triangle triangle){
			if (right != null)	System.out.println("Error in Edge.set_right");
			right = triangle;
		}
		public void remove(Triangle triangle){
			if (left == triangle) left = null;
			if (right == triangle) right = null;
		}
		public boolean contains(Vertex v){
			return (v == start || v == end);
		}
		public Vertex get_common_vertex(Edge edge){
			if (edge.contains(start)) return start;
			if (edge.contains(end)) return end;
			return null;
		}
	}
	class Triangle {
		Edge[] edges = new Edge[3]; 
		Triangle(Edge e0, Edge e1, Edge e2){
			edges[0] = e0;
			edges[1] = e1;
			edges[2] = e2;
			
			if (e1.contains(e0.end))	e0.set_left(this);
			else						e0.set_right(this);
			if (e2.contains(e1.end))	e1.set_left(this);
			else						e1.set_right(this);
			if (e0.contains(e2.end))	e2.set_left(this);
			else						e2.set_right(this);
		}
		public void delete(){
			edges[0].remove(this);
			edges[1].remove(this);
			edges[2].remove(this);
		}
		public Vertex get_the_other_vertex(Edge edge){
			if (edge == edges[0])	return edges[1].get_common_vertex(edges[2]);
			if (edge == edges[1])	return edges[2].get_common_vertex(edges[0]);
			if (edge == edges[2])	return edges[0].get_common_vertex(edges[1]);
			return null;
		}
		public Edge get_the_other_edge(Vertex v){
			if (!edges[0].contains(v)) return edges[0];
			if (!edges[1].contains(v)) return edges[1];
			if (!edges[2].contains(v)) return edges[2];
			return null;
		}
		public Vertex get_vertex(int i){
			if (i==0)	return edges[1].get_common_vertex(edges[2]);
			if (i==1)	return edges[2].get_common_vertex(edges[0]);
			if (i==2)	return edges[0].get_common_vertex(edges[1]);
			return null;
		}
	}
	
}