package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class BridgeHelper extends PlaneHelper
{
	public Suggestion makeSuggestion(Chateau chateau){
		Line currentLine = chateau.currentLine();

		
		int size = chateau.currentLines.size();
		if (size == 0 || size-(size/2)*2 != 0)
			return null;

		Vector connected0 = currentLine.gatherConnectedLines(chateau.currentLines);
		if (connected0.size() != size/2)
			return null;
		
		Hashtable hash = new Hashtable();

		if (connected0.size() == 1){
			// simple bridge
			Line a = (Line) connected0.elementAt(0);
			Line b = (Line) chateau.currentLines.elementAt(0);

			// same plane test
			if (!a.onSamePlane(b))
				return null;
			if (a.on_line(b.start) || a.on_line(b.end) ||
				b.on_line(a.start) || b.on_line(a.end))
					return null;
			if (a.getCrossPoint(b) != null)	//a.cross(b))
				return null;
			
						
			/*
			if (!a.parallel(b)){
				Vector3 v0 = Vector3.subtract(a.start, b.start);
				Vector3 v1 = Vector3.subtract(a.end, b.end);
				Vector3 u0 = Vector3.subtract(a.start, b.end);
				Vector3 u1 = Vector3.subtract(a.end, b.start);
				if (!Line.parallel(v0,v1) && !Line.parallel(u0,u1))
					return null;
				if (Math.abs(a.getLength()-b.getLength())>Chateau.ERROR)
					return null;
			}
			*/
			
			
			//already on a plate (TrimPlateHelper)
			for(int i=0; i<a.plates.size(); i++){
				Plate plate = (Plate) a.plates.elementAt(i);
				if (plate.on_plate(b))
					return null;
			}
			for(int i=0; i<b.plates.size(); i++){
				Plate plate = (Plate) b.plates.elementAt(i);
				if (plate.on_plate(a))
					return null;
			}

			
			hash.put(a,b);
		}
		else {
		
		Plane plane0 = find_common_plane(connected0);
		if (plane0 == null) 
			return null;


		//---
		
		Line line1 = null;
		for(int i=0; i<chateau.currentLines.size(); i++){
			Line line = (Line)chateau.currentLines.elementAt(i);
			if (!connected0.contains(line)){
				line1 = line;
				break;
			}
		}
		if (line1 == null)
			return null;
		
		Vector connected1 = line1.gatherConnectedLines(chateau.currentLines);
		
		if (connected1.size() != connected0.size())
			return null;

		Plane plane1 = find_common_plane(connected1);
		if (plane1 == null || !plane0.parallel(plane1) || plane0.equivalent(plane1)) 
			return null;
		
			
		
		//---
     	/*
		Vector mapped0 = map(connected0, plane0);
		Vector mapped1 = map(connected1, plane0);
		
		for(int i=0; i<mapped0.size(); i++){
			Line line = (Line) mapped0.elementAt(i);
			Line corresponding_line = find_corresponding_line(mapped1, line);
			if (corresponding_line == null)
				return null;
			hash.put(connected0.elementAt(i), connected1.elementAt(mapped1.indexOf(corresponding_line)));
		}
		*/
		Vector remaining_connected1 = Util.duplicateVector(connected1);
		for(int i=0; i<connected0.size(); i++){
			Line line0 = (Line) connected0.elementAt(i);
			Line line2 = find_closest_parallel_line(remaining_connected1, line0);
			if (line2 == null)
				return null;
			remaining_connected1.removeElement(line2);
			hash.put(line0, line2);
		}
		
		//---



		}



		// let us make a suggestion!!!
			
		World originalWorld = 	new World(chateau);

		for(int i=0; i<connected0.size(); i++){
			Line a = (Line) connected0.elementAt(i);
			Line b = (Line) hash.get(a);
			
			Line line0;
			Line line1;
			if (Vector3.dotProduct(a.getVector3(), b.getVector3())>0){
				line0 = chateau.addLine(new Line(a.startJoint, b.startJoint));
				line1 = chateau.addLine(new Line(a.endJoint, b.endJoint));
			}
			else {
				line0 = chateau.addLine(new Line(a.startJoint, b.endJoint));
				line1 = chateau.addLine(new Line(a.endJoint, b.startJoint));
			}
			
			chateau.addCurrentLine(line0);
			chateau.addCurrentLine(line1);
			
			chateau.addPlate(new Plate(a, line0, b, line1));
		
		}
		
		Suggestion suggestion = new Suggestion(chateau, "Bridge");
		chateau.restoreWorld(originalWorld);
		return suggestion;
	}
	
	
	
	
	public Vector map(Vector lines, Plane plane){
		Vector mapped = new Vector();
		double min_x = 1000000;
		double min_z = 1000000;
		
		for(int i=0; i<lines.size(); i++){
			Line line = plane.map((Line) lines.elementAt(i));
			mapped.addElement(line);	// z=0
			
			min_x = Math.min(line.start.x, min_x);
			min_z = Math.min(line.start.z, min_z);
	
			min_x = Math.min(line.end.x, min_x);
			min_z = Math.min(line.end.z, min_z);
		}
		
		for(int i=0; i<mapped.size(); i++){
			Line line = (Line) mapped.elementAt(i);
			line.start = new Vector3(line.start.x - min_x, 0, line.start.z - min_z);	
			line.end = new Vector3(line.end.x - min_x, 0, line.end.z - min_z);	
			
		}
		
		return mapped;
			
	
	}
	
	public Line find_corresponding_line(Vector mapped, Line line){
		for(int i=0; i<mapped.size(); i++){
			Line l = (Line) mapped.elementAt(i);
			if (l.equivalent(line))
				return l;
		}
		return null;
	
	}
	public Line find_closest_parallel_line(Vector lines, Line line){
		double min = 100000;
		Line closest = null;
		for(int i=0; i<lines.size(); i++){
			Line l = (Line) lines.elementAt(i);
			if (line.parallel(l)){
				double d = line.parallel_distance(l);
				if (d<min){
					min = d;
					closest = l;
				}
			}
		}
		
		return closest;
	
	}
}
