package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;

public class Chateau implements java.awt.event.MouseListener, 
								java.awt.event.MouseMotionListener, 
								java.awt.event.KeyListener
{
	public static final double ERROR = +1.0E-3;

	public final java.awt.Color SUGGESTION_COLOR = new java.awt.Color(234, 148, 255)	;
	




	
	
	
	
	//
	// Undo (History management)
	//
	
	Vector worlds = new Vector();
	public void pushUndo(){
		if (worlds.size() == 0 || ((World) worlds.elementAt(worlds.size()-1)).lines.size() != lines.size() || 
								  ((World) worlds.elementAt(worlds.size()-1)).currentLines.size() != currentLines.size() )			
			worlds.addElement(new World(this));	
	}
	public void popUndo(){
		World world = (World) worlds.elementAt(worlds.size()-1);
		worlds.removeElement(world);
		restoreWorld(world);
	}
	public void cancelPushUndo(){
		worlds.removeElementAt(worlds.size()-1);
	}
	public void undo(){
		if (worlds.size()>0){
			PlaySound.play("sounds\\lock.wav");//star.wav");
		
			popUndo();
			predict_and_suggest();	
			
		}
	}

	
	
	
	
	
	public void restoreWorld(World world){

		Vector old_vehicles = get_vehicles();
		clear_scene_main();

		joints = world.joints;
		lines = world.lines;
		planes = world.planes;
		plates = world.plates;

		currentLines = world.currentLines;
		currentPlane = world.currentPlane;

		Vector new_vehicles = get_vehicles();

		for(int i=0; i<new_vehicles.size(); i++){
			Transformable vehicle = (Transformable)new_vehicles.elementAt(i);
			if (!old_vehicles.contains(vehicle))
				vehicle.setParent(jaliceFrame.scene);
		}
		
		for(int i=0; i<old_vehicles.size(); i++){
			Transformable vehicle = (Transformable)old_vehicles.elementAt(i);
			if (!new_vehicles.contains(vehicle))
				vehicle.setParent(null);
		}
		
		
		
		
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			line.restoreStyle();
		}
		for(int i=0; i<planes.size(); i++){
			Plane plane = (Plane) planes.elementAt(i);
			plane.show();
			

				
			if (plane.is_ground)
				ground = plane;
		}

		
		for(int i=0; i<currentLines.size(); i++)
			((Line)currentLines.elementAt(i)).activate();
			
	}







	
	
	
	
	
	
	//
	//	File (load/save)
	//
	
	String dirname = ".\\models\\";
	String filename = "";
	public void file(){
		FileControl fileControl = new FileControl(this);
		fileControl.show();
		switch(fileControl.command){
			case FileControl.LOAD:
				load();
				break;
			case FileControl.SAVE:
				save();
				break;
			case FileControl.EXPORT:
				export();
				break;
		}
	}
	public void load(){
		fileSub("Open Model", java.awt.FileDialog.LOAD);
		
		if( filename!=null )
		{
			System.out.println("load "+filename);
			Vector[] vectors = FileHandler.load(dirname+filename);//World world = 
			loadWorld(vectors[0], vectors[1], vectors[2]);
		}
	}
	public void save(){
		fileSub("Save Model", java.awt.FileDialog.SAVE);
		
		if( filename!=null )
		{
			FileHandler.save(new World(this), dirname+filename);
		}
	}
	public void export(){
		fileSub("Export Obj", java.awt.FileDialog.SAVE);
		
		if( filename!=null )
		{
			System.out.println("export is not implemented yet"+filename);
			
		}
	}
	
	
	public void fileSub(String title, int type){
		java.awt.FileDialog fileDialog = new java.awt.FileDialog(jaliceFrame, "Open Model", type);
		fileDialog.setDirectory( dirname );
		fileDialog.setFile( "*.cht" );
		centerDialog(fileDialog);
		fileDialog.show();

		dirname = fileDialog.getDirectory();
		filename = fileDialog.getFile();
	}
	public void centerDialog(java.awt.FileDialog fileDialog){
		java.awt.Rectangle b = jaliceFrame.getBounds();
		java.awt.Dimension s = new java.awt.Dimension(506, 250);
		fileDialog.setLocation(b.x+b.width/2-s.width/2, b.y+b.height/2-s.height/2);
	}


	public void loadWorld(Vector _joints, Vector _lines, Vector _plates){
		pushUndo();
		init_scene();
	
		for(int i=0; i<_joints.size(); i++){
			double[] xyz = (double[]) _joints.elementAt(i);
			Joint joint = new Joint(new Vector3(xyz[0], xyz[1], xyz[2]));
			joints.addElement(joint);
			joint.addToScenegraph(jaliceFrame.scene);
		}
		for(int i=0; i<_lines.size(); i++){
			int[] pq = (int[]) _lines.elementAt(i);
			Joint p = (Joint)joints.elementAt(pq[0]);
			Joint q = (Joint)joints.elementAt(pq[1]);
			Line line = new Line(p, q);
			addLine_main(line);

		}
		for(int i=0; i<_plates.size(); i++){
			Vector _ls = (Vector) _plates.elementAt(i);
			Vector ls = new Vector();
			for(int j=0; j<_ls.size(); j++){
				int index = ((Integer)_ls.elementAt(j)).intValue();
				ls.addElement((Line)lines.elementAt(index));
			}
			Plate plate = new Plate(ls);
			addPlate_main(plate);
		}
		
	}























	// from command menu
	public void init(){
		pushUndo();
		init_scene();
	}

	public void init_scene(){
	
		clearSuggestions();
		clear_scene();

		createGround();
		ground.show();
		
		createCenterBall();
		hideCenterBall();
		
		planes.addElement(ground);

		jaliceFrame.init_camera_position();

		//currentLine = null;
		clearCurrentLines();
		currentPlane = ground;
	}
	public void clear_scene(){
		
		removeFromScenegraph();
		
		clear_scene_main();
	}
	
	public void clear_scene_main(){

		lines = new Vector();
		joints = new Vector();
		plates = new Vector();
		planes = new Vector();
		
		for(int j=0; j<candidate_lines.size(); j++){
			Line line = (Line)candidate_lines.elementAt(j);
			line.delete();
		}
		candidate_lines = new Vector();

		//visible_planes = new Vector();

	}

	public void removeFromScenegraph(){

		for(int i=0; i<joints.size(); i++){
			Joint joint = (Joint) joints.elementAt(i);
			joint.removeFromScenegraph();
		}
		for(int j=0; j<lines.size(); j++){
			Line line = (Line)lines.elementAt(j);
			line.removeFromScenegraph();
		}
		for(int j=0; j<plates.size(); j++){
			Plate plate = (Plate)plates.elementAt(j);
			plate.removeFromScenegraph();
		}
		for(int i=0; i<planes.size(); i++){
			Plane plane = (Plane)planes.elementAt(i);
			plane.removeFromScenegraph();
		}
		
		if (centerBall != null)
			centerBall.removeFromScenegraph();
	}


	public Vector get_vehicles(){
		Vector elements  = new Vector();
		Util.appendVector(elements, joints);
		Util.appendVector(elements, lines);
		Util.appendVector(elements, plates);
		Util.appendVector(elements, planes);
		
		Vector vehicles = new Vector();
		for(int i=0; i<elements.size(); i++){
			ChateauElement e = (ChateauElement)elements.elementAt(i);
			vehicles.addElement(e.getVehicle());
		}
		
		return vehicles;
	}



	public Vector stroke = new Vector();
	public void add_point(Point p){
		stroke.addElement(p);
	}
	public void clear_stroke(){
		stroke = new Vector();
	}
	//Overlay overlay;
	public void paintOverlay(Graphics g){

		if (snap != null)
			snap.paint(g);	
		
		if (operation_status == NONE || operation_status == SUGGESTION)
			paintSuggestions(g);
	}
	
	

	
	
	
	
	
	
	
	
	
	
	

	
	//
	// Scene description
	//
	
	static Vector joints = new Vector(); 		
	static Vector lines = new Vector();
	static Vector plates = new Vector();
	static Vector planes = new Vector();

	Plane currentPlane;
	
	Vector candidate_lines = new Vector();
	Line primary_candidate_line;

	

	
	
	
	
	
	
	
	
	





	
	//
	// Suggestions
	//

	static Vector suggestions = new Vector();
	public void paintSuggestions(Graphics g){
		if (currentSuggestion != null){
			java.awt.Dimension size = jaliceFrame.awtComponent.getSize();
			g.setColor(SUGGESTION_COLOR);//java.awt.Color.red);
			g.fillRect(0,0,size.width, Suggestion.PREVIEW_FRAME);
			g.fillRect(0,size.height-Suggestion.PREVIEW_FRAME,size.width, Suggestion.PREVIEW_FRAME);

			g.fillRect(0,0,Suggestion.PREVIEW_FRAME, size.height);
			g.fillRect(size.width-Suggestion.PREVIEW_FRAME,0,Suggestion.PREVIEW_FRAME, size.height);
		}
			
		//	g.drawImage(currentSuggestion.fullImage, 0,0,null);
	
		for(int i=0; i<suggestions.size(); i++){
			Suggestion suggestion = (Suggestion) suggestions.elementAt(i);
			java.awt.Rectangle bbox = suggestion.getBoundingBox();
			//g.drawImage(suggestion.getThumbnailImage(), bbox.x, bbox.y, null);
			
			if (suggestion == currentSuggestion){
				g.setColor(SUGGESTION_COLOR);//java.awt.Color.red);
				g.drawRect(bbox.x, bbox.y, bbox.width-1, bbox.height-1);
				g.drawRect(bbox.x+1, bbox.y+1, bbox.width-3, bbox.height-3);
			}
		}
	}
	public void setSuggestionsBoundingBox(){
		java.awt.Dimension size = jaliceFrame.awtComponent.getSize();
		for(int i=0; i<suggestions.size(); i++){
			Suggestion suggestion = (Suggestion) suggestions.elementAt(i);
			java.awt.Rectangle bbox = new java.awt.Rectangle(i*Suggestion.WIDTH, size.height - Suggestion.HEIGHT, Suggestion.WIDTH, Suggestion.HEIGHT);
			suggestion.setBoundingBox(bbox, i);
		}	
	}
	
	Suggestion currentSuggestion = null;
	Suggestion prevSuggestion = null;
	World currentWorld;
	public boolean pressSuggestion(Point p){
		pushUndo();
		
		Suggestion suggestion = null;
		for(int i=0; i<suggestions.size(); i++){
			Suggestion s = (Suggestion) suggestions.elementAt(i);
			java.awt.Rectangle bbox = s.getBoundingBox();
			if (bbox.contains(p))
				suggestion = s;
		}
		currentSuggestion = suggestion;

		if (suggestion == null)
			return false;
			
		currentWorld = new World(this);
					
		return true;

	}
	public void finishSuggestion(Point p){
		if (currentSuggestion == null){
			cancelPushUndo();
			return;
		}
		else
			PlaySound.play("sounds\\lock.wav");
			
		//currentSuggestion.apply(this);
		//suggestions = new Vector();
		clearSuggestions();
		
		currentSuggestion = null;

		predict_and_suggest();
		
		repaint();
	}
	public void repaint(){
		jaliceFrame.awtComponent.repaint();
	}
	public void dragSuggestion(Point p){
		Suggestion suggestion = null;
		for(int i=0; i<suggestions.size(); i++){
			Suggestion s = (Suggestion) suggestions.elementAt(i);
			java.awt.Rectangle bbox = s.getBoundingBox();
			if (bbox.contains(p))
				suggestion = s;
		}
		currentSuggestion = suggestion;
		
	
		
		if (prevSuggestion != currentSuggestion){
			PlaySound.play("sounds\\kasha.wav");
			if (currentSuggestion == null)
				restoreWorld(currentWorld);
			else {
				restoreWorld(currentSuggestion.world);
			}
		}
		
		
		prevSuggestion = currentSuggestion;
	}
	
	
	
	
	
	
	

	public Vector helpers = new Vector();	// registerd by HelperCenter.java
	public void makeSuggestions(){
		clearSuggestions();
		
		jaliceFrame.disableUpdate();
		for(int i=0; i<helpers.size(); i++){
			Helper helper = (Helper) helpers.elementAt(i);
			Suggestion suggestion = helper.makeSuggestion(this);
			if (suggestion != null)
				suggestions.addElement(suggestion);
		}
		jaliceFrame.enableUpdate();


		setSuggestionsBoundingBox();
	}
	public void clearSuggestions(){
		for(int i=0; i<suggestions.size(); i++){
			Suggestion suggestion = (Suggestion)suggestions.get(i);
			suggestion.vehicle.setParent(null);
		}
		suggestions = new Vector();
	}	
	
	
	
	
	
	
	
	

	
	
	
	
	
	
	
	
	
	
	
	public Plane pickPlane(Point p){
		Vector planeVisuals = new Vector();
		for(int i=0; i<planes.size(); i++)
			planeVisuals.addElement(((Plane)planes.elementAt(i)).planeVisual);

		pickedVisual = jaliceFrame.pickVisual(p, planeVisuals);

		for(int i=0; i<planes.size(); i++)
			if (pickedVisual == ((Plane)planes.elementAt(i)).planeVisual)
				return (Plane)planes.elementAt(i);
		return null;
	}
	public Joint pickJoint(Point p){
		Vector jointVisuals = new Vector();
		for(int i=0; i<joints.size(); i++)
			jointVisuals.addElement(((Joint)joints.elementAt(i)).jointVisual);

		pickedVisual = jaliceFrame.pickVisual(p, jointVisuals);

		for(int i=0; i<joints.size(); i++)
			if (pickedVisual == ((Joint)joints.elementAt(i)).jointVisual)
				return (Joint)joints.elementAt(i);
		return null;
	}

	public Object pickJointPlatePlane(Point p){
		Vector visuals = new Vector();
		Hashtable hash = new Hashtable();
		
		for(int i=0; i<joints.size(); i++){
			Joint joint = (Joint)joints.elementAt(i);
			Visual visual = joint.jointVisual;
			visuals.addElement(visual);
			hash.put(visual, joint);
		}
		for(int i=0; i<plates.size(); i++){
			Plate plate = (Plate)plates.elementAt(i);
			if (!plate_is_containd_in_visibleplanes(plate)){
				Visual visual = plate.plateVisual;
				visuals.addElement(visual);
				hash.put(visual, plate);
			}
		}
		for(int i=0; i<planes.size(); i++){
			Plane plane = (Plane)planes.elementAt(i);
			Visual visual = plane.planeVisual;
			visuals.addElement(visual);
			hash.put(visual, plane);
		}

		pickedVisual = jaliceFrame.pickVisual(p, visuals);

		if (pickedVisual == null)
			return null;
			
		return hash.get(pickedVisual);
		
	}
	public boolean plate_is_containd_in_visibleplanes(Plate plate){
		Plane p = plate.getPlane();
		for(int i=0; i<planes.size(); i++){
			Plane plane = (Plane)planes.elementAt(i);
			if (plane.equivalent(p)){
				return true;
			}
		}
		return false;
		
	}
	
	public Object pickPlatePlane(Point p){
		Object picked_object = null;
		double min = Double.MAX_VALUE;
		
		Vector3 cameraPt = jaliceFrame.cameraVehicle.getPosition(jaliceFrame.scene);
		
		InfiniteLine line = jaliceFrame.get_camera_lay(p);
		for(int i=0; i<plates.size(); i++){
			Plate plate = (Plate)plates.elementAt(i);
			Vector3 v = plate.get_cross_point(line);
			if (v!= null){
				double d = Vector3.subtract(cameraPt, v).getLength();
				if (d < min){
					picked_object = plate;
					min = d;
				}
			}
		}
		for(int i=0; i<planes.size(); i++){
		Plane plane = (Plane)planes.elementAt(i);
			Vector3 v = plane.get_cross_point(line);
			if (v!= null){
				double d = Vector3.subtract(cameraPt, v).getLength();
				if (d < min){
					picked_object = plane;
					min = d;
				}
			}
		}
		
		return picked_object;
	}
//	public Object pickPlatePlane(Point p){
//		Vector visuals = new Vector();
//		Hashtable hash = new Hashtable();
//		for(int i=0; i<plates.size(); i++){
//			Plate plate = (Plate)plates.elementAt(i);
//			if (!plate_is_containd_in_visibleplanes(plate)){
//				Visual visual = plate.plateVisual;
//				visuals.addElement(visual);
//				hash.put(visual, plate);
//			}
//		}
//		for(int i=0; i<planes.size(); i++){
//			Plane plane = (Plane)planes.elementAt(i);
//			Visual visual = plane.planeVisual;
//			visuals.addElement(visual);
//			hash.put(visual, plane);
//		}
//
//		pickedVisual = jaliceFrame.pickVisual(p, visuals);
//
//		if (pickedVisual == null)
//			return null;
//			
//		return hash.get(pickedVisual);
//		
//	}
	public Object pickLinePlate(Point p){
		Vector visuals = new Vector();
		Hashtable hash = new Hashtable();
		for(int i=0; i<plates.size(); i++){
			Plate plate = (Plate)plates.elementAt(i);
			Visual visual = plate.plateVisual;
			visuals.addElement(visual);
			hash.put(visual, plate);
		}
		for(int i=0; i<lines.size(); i++){
			Line line = (Line)lines.elementAt(i);
			Visual visual = line.lineVisual;
			visuals.addElement(visual);
			hash.put(visual, line);
		}
		for(int i=0; i<candidate_lines.size(); i++){
			Line line = (Line)candidate_lines.elementAt(i);
			Visual visual = line.lineVisual;
			visuals.addElement(visual);
			hash.put(visual, line);
		}

		pickedVisual = jaliceFrame.pickVisual(p, visuals);

		if (pickedVisual == null)
			return null;
			
		return hash.get(pickedVisual);
		
	}

	
	
	
	public Plane getTargetPlane(){
		return  currentLine().getNextPlane(currentPlane);	// this may call getPlane()
	}
	public void switchPlane(Plane nextPlane){

		
		if (!currentPlane.is_ground){
			planes.removeElement(currentPlane);
			currentPlane.delete();
		}
		if (!nextPlane.is_ground){
			planes.addElement(nextPlane);
			nextPlane.addToScenegraph(jaliceFrame.scene);
			nextPlane.show();
			addLinesToPlane(nextPlane);
		}
		
		currentPlane = nextPlane;
		
	}
	public void addLinesToPlane(Plane plane){
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (plane.contains(line))
				plane.lines.addElement(line);
		}
	}	

	// from start_drawing()
	public Plane getDrawingPlane(Plane plane){
		for(int i=0; i<planes.size(); i++){
			Plane p = (Plane) planes.elementAt(i);
			if (plane.equivalent(p))
				return p;
		}
		addLinesToPlane(plane);
		return plane;
	}




	public boolean all_lines_are_selected(Vector lines){
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (!currentLines.contains(line))
				return false;
		}
		return true;
	}


	public void click(Point p, java.awt.event.MouseEvent event){
		Object object = pickLinePlate(p);
		
		//Line line = pickLine(p);
		if (object instanceof Plate){
			flush_candidates();//event);

			Plate plate = (Plate) object;
			
			if (event.getClickCount() == 2){
				Line line = (Line) plate.lines.elementAt(0);
				Vector lines = line.gatherConnectedLines();
				for(int i=0; i<lines.size(); i++)
					addCurrentLine((Line)lines.elementAt(i));
			}
			else {
				if (all_lines_are_selected(plate.lines)){
					for(int i=0; i<plate.lines.size(); i++){
						Line line = (Line) plate.lines.elementAt(i);
						removeCurrentLine(line);
					}
				}
				else {
					for(int i=0; i<plate.lines.size(); i++){
						Line line = (Line) plate.lines.elementAt(i);
						addCurrentLine(line);
					}
				}
			}
			//Plane plane = getPlane(plate.getPlane());
			//switchPlane(plane);
						
			predict_and_suggest();
			return;
			//}else object = null;
		}
		
		Line line = (Line) object;
		if (candidate_lines.size()>0 && line == null){
			// click nothing
			// finalize the primary choise
			if (primary_candidate_line != null){
				flush_candidates();
				predict_and_suggest();
			}
			else {
				// abort from prediction
				
				flush_candidates();
				disactivateAll();
				
			}
				
		}
		else if (candidate_lines.size()>0 && candidate_lines.contains(line)){
			if (primary_candidate_line != null) {
				// select beautified candidate
				if (primary_candidate_line != line){
					primary_candidate_line.set_candidate_style();
					primary_candidate_line = line;
					primary_candidate_line.set_primary_candidate_style();
				}
				else {
					flush_candidates();
					predict_and_suggest();
				
				}
			}
			else {
				// select predicted candidate
				primary_candidate_line = line;
				flush_candidates();
				predict_and_suggest();
			}
				
		}
		else if (line != null){
			// click on line (not candidate)
			
			flush_candidates();

			pushUndo();
								
			if (event.getClickCount() == 2){

				Vector lines = line.gatherConnectedLines();
				for(int i=0; i<lines.size(); i++)
					addCurrentLine((Line)lines.elementAt(i));

				
				addCurrentLine(line);	// bring to the end!
			
			}
			else {
				if (currentLines.contains(line)){
					removeCurrentLine(line);
				}
				else {
					addCurrentLine(line);
				}
			}

			predict_and_suggest();


		}	
		else if (line == null){
			disactivateAll();
		}
	
	}
	public void disactivateAll(){
			pushUndo();

			clearCurrentLines();
			clearSuggestions();
			switchPlane(ground);
	}
	public void clickFrame(){
			//System.out.println("Click");
			flush_candidates();
			disactivateAll();
	}





	Vector currentLines = new Vector();
	Vector prevCurrentLines = new Vector();
	public void clearCurrentLines(){
		for(int i=0; i<currentLines.size(); i++)
			((Line)currentLines.elementAt(i)).disactivate();
		prevCurrentLines = currentLines;
		currentLines = new Vector();
	}
	public void addCurrentLine(Line line){
		if (currentLines.contains(line))
			currentLines.removeElement(line);	//bring to end;
			//return;
		
		currentLines.addElement(line);
		line.activate();
	}
	public void removeCurrentLine(Line line){
		currentLines.removeElement(line);
		line.disactivate();
	}
	public Line currentLine(){
		if (currentLines.size()==0)
			return null;
		else
			return (Line) currentLines.elementAt(currentLines.size()-1);
	}
	
	
	
	


	public void erase(Point p){
		flush_candidates();

		pushUndo();


		Object picked = pickLinePlate(p);
		
		if (picked instanceof Line)
			eraseLine((Line) picked);
		else if (picked instanceof Plate)
			erasePlate((Plate) picked);
			
		//jaliceFrame.renderTarget.onscreenUpdate();//.updateOnscreenBuffer(); //render();
		jaliceFrame.renderTarget.clearAndRenderOffscreen();
	}
	// from Controls
	public void erase(){
		clearSuggestions();
		flush_candidates();

		if (currentLines.size() == 0)
			return;
	
		Vector l = Util.duplicateVector(currentLines);

		pushUndo();
		PlaySound.play("sounds\\po.wav");
		
		Vector p = Util.duplicateVector(plates);

		for(int i=0; i<p.size(); i++){
			Plate plate = (Plate) p.elementAt(i);
			if (Util.isSubsetVector(plate.lines, l))
				erasePlate(plate);
		}
		for(int i=0; i<l.size(); i++){
			Line line = (Line)l.elementAt(i);
			if (lines.contains(line))
				eraseLine(line);
		}
		
	}
	
	public void eraseLine(Line line){
		//if (line.plates.size() == 0)
			deleteLine(line);
	}
	public void erasePlate(Plate plate){

		deletePlate(plate);

	}

		
	
	public void flush_candidates(){		//java.awt.event.InputEvent event){
		//clearSuggestions();	
		if (candidate_lines.size() == 0)
			return;
		for(int i=0; i<candidate_lines.size(); i++){
			Line line = (Line) candidate_lines.elementAt(i);
			//if (line != primary_candidate_line){
				line.delete();
				lines.removeElement(line);
			//}
		}	
		candidate_lines = new Vector();

			
		if (primary_candidate_line == null)
			return;
			
		pushUndo();
		
		Line line = primary_candidate_line;
		primary_candidate_line = null;
		line.set_regular_style();

		Joint startJoint = findJoint(line.start);
		Joint endJoint = findJoint(line.end);
		//line.setJoints(startJoint, endJoint);
		
		Line newLine = new Line(startJoint, endJoint);
		newLine = addLine(newLine);

		addCurrentLine(newLine);		          
		
		
	}
	public Joint findJoint(Vector3 p){
		for(int i=0; i<joints.size(); i++){
			Joint joint = (Joint)joints.elementAt(i);
			if (almost_identical(p, joint.position))
				return joint;
		}
		Joint joint = new Joint(p);
		joints.addElement(joint);
        joint.addToScenegraph(jaliceFrame.scene);

		return joint;
	}
	public boolean almost_identical(Vector3 p0, Vector3 p1){
		return (Vector3.subtract(p1,p0).getLength()<0.0001);
	}
	
	
	
	public void predict_and_suggest(){
		
		//pushUndo();

		if (suggestion_switch)
			makeSuggestions();
		if (prediction_switch)
			makePredictions();
	}

	public void makePredictions(){

		Vector predicted_lines = Predictor.predict(currentLines);
		if (predicted_lines == null)
			return;
	
		candidate_lines = new Vector();	
		for(int i=0; i<predicted_lines.size(); i++){
			Line line = (Line) predicted_lines.elementAt(i);
			if (!exisiting_line(line)){
					
				candidate_lines.addElement(line);
				line.addToScenegraph(jaliceFrame.scene);
				line.set_candidate_style();
			}
		}
		
	}
	public boolean exisiting_line(Line line){	
		for(int i=0; i<lines.size(); i++)
			if (((Line)lines.elementAt(i)).equivalent(line))
				return true;
		return false;
	}





	public void connectJoints(Joint startJoint, Joint endJoint){
		if (startJoint == endJoint)
			return;
		
		pushUndo();

		Line newLine = new Line(startJoint, endJoint);
		newLine = addLine(newLine);
			
		newLine.set_regular_style();
		
		//currentLine.activate();
		addCurrentLine(newLine);
		
		predict_and_suggest();
	}

	public Line addLine(Line line){
		Line l = line.startJoint.getCommonLine(line.endJoint);
		if (l != null)
			return l;
		
		return addLine_main(line);
	}
	public Line addLine_main(Line line){
	
		lines.addElement(line);
		line.addToScenegraph(jaliceFrame.scene);


		line.startJoint.lines.addElement(line);
		line.endJoint.lines.addElement(line);

		
		for(int i=0; i<planes.size(); i++){
			Plane plane = (Plane) planes.elementAt(i);
			if (plane.contains(line))
				plane.lines.addElement(line);//line.addParentPlane(plane);	// this includes "parentPlane.addElement(this)"
		}
		
		
		
		return line;
	}
	public Plate addPlate(Plate plate){
		for(int i=0; i<plates.size(); i++){
			Plate l = (Plate) plates.elementAt(i);
			if (l.equivalent(plate))
				return l;
		}
		return addPlate_main(plate);
	}
	public Plate addPlate_main(Plate plate){
		plates.addElement(plate);
		plate.addToScenegraph(jaliceFrame.scene);
	
		for(int i=0; i<plate.lines.size(); i++){
			Line line = (Line) plate.lines.elementAt(i);
			line.addPlate(plate);
		}
		
		return plate;
	}

	public void deleteLine(Line line){
		if (line.plates.size() != 0)
			return;
		
		if (currentLines.contains(line))
			currentLines.removeElement(line);
		
		System.out.println("planes.size() "+planes.size());
		for(int i=0; i<planes.size(); i++){
			Plane plane = (Plane) planes.elementAt(i);
			if (plane.contains(line))
				System.out.println("removeElement "+line);
			plane.lines.removeElement(line);
		}
		line.delete();
		lines.removeElement(line);
	}

	public void deletePlate(Plate plate){	
		plate.delete();
		plates.removeElement(plate);
	}


	
	
	
	


	public Line pickLine(Point p){
		Vector lineVisuals = new Vector();
		for(int i=0; i<lines.size(); i++){
			Line line = (Line)lines.elementAt(i);
			lineVisuals.addElement(line.lineVisual);
		}
		for(int i=0; i<candidate_lines.size(); i++){
			Line line = (Line)candidate_lines.elementAt(i);
			lineVisuals.addElement(line.lineVisual);
		}

		pickedVisual = jaliceFrame.pickVisual(p, lineVisuals);

		for(int i=0; i<lines.size(); i++){
			Line line = (Line)lines.elementAt(i);
			if (pickedVisual == line.lineVisual)
					return line;
		}
		for(int i=0; i<candidate_lines.size(); i++){
			Line line = (Line)candidate_lines.elementAt(i);
			if (pickedVisual == line.lineVisual)
					return line;
		}
		return null;
	}


















	//
	// Drawing a new line by left-button-dragging
	//

	Line newLine;
	Projectable currentProjectable;
	Plane basePlane;
	Point startPoint;
	Vector3 startPosition;
	Point prevPoint;
	Snap snap;
	double zigzagCount = 0;
	public void start_drawing(Point p){
		startPoint = p;
		clearSuggestions();
		
		Object pickedStart = pickPlatePlane(p);
			
		if (pickedStart == null)
			return;
		
		Visual pickedVisual;
		if (pickedStart instanceof Plate){
			Plate plate = (Plate) pickedStart;
			pickedVisual = plate.plateVisual;
			basePlane = getDrawingPlane(plate.getPlane());
		}
		else if (pickedStart instanceof Plane){
			basePlane = (Plane) pickedStart;
			pickedVisual = basePlane.planeVisual;
		}
		else
			return;



		PlaySound.play("sounds\\click.wav");
		
		//currentProjectable = new Projectable(basePlane.planeVehicle, basePlane.planeVisual, basePlane.lines,basePlane.base, basePlane.normal);
		currentProjectable = new Projectable(basePlane.base, basePlane.normal);

		//double distance_to_pickpoint = jaliceFrame.get_pick_depth(p, pickedVisual);//2,10,30
		Vector3 pick_position = currentProjectable.project(p);		
		double distance_to_pickpoint = Vector3.subtract(pick_position, jaliceFrame.cameraVehicle.getPosition(null)).getLength();
		
		snap = new Snap(basePlane, this, distance_to_pickpoint);
		
		startPosition = currentProjectable.project(p);
		
		startPosition = snap.snap(startPosition);
		snap.prepare_guides_after_start(startPosition);
		
		newLine = new Line(startPosition, startPosition);
		newLine.addToScenegraph(jaliceFrame.scene);
		newLine.activate();
		
		zigzagCount = 0;
		prevPoint = startPoint;

	}

	public void drawing(Point p){
		teddy.Vector2 a = new teddy.Vector2(startPoint, prevPoint);
		teddy.Vector2 b = new teddy.Vector2(startPoint, p);
		if (!startPoint.equals(prevPoint) && a.get_relative_angle(b) > Math.PI/2)
			zigzagCount++;
		if (zigzagCount>=3){
			stopDrawingAndErase();
			return;
		}
		prevPoint = p;
		
		if (teddy.Vector2.distance(p, startPoint) > 10){
			flush_candidates();
			if (snap != null)
			snap.enable_feedback();
		}
			
	
		Vector3 end = currentProjectable.project(p);

		if (snap != null)
		end = snap.snap(end);

		newLine.removeFromScenegraph();
		newLine = new Line(startPosition, end);
		newLine.addToScenegraph(jaliceFrame.scene);
		newLine.activate();


		
	}

	public void finish_drawing(Point p, java.awt.event.MouseEvent event){
		snap = null;
		
		newLine.removeFromScenegraph();


		if (teddy.Vector2.distance(startPoint, p)<10){
			click(startPoint, event);
			return;
		}

		PlaySound.play("sounds\\click.wav");

		Joint startJoint = findJoint(newLine.start);
		Joint endJoint = findJoint(newLine.end);
		
		Line line = new Line(startJoint, endJoint);
		line = addLine(line);

		addCurrentLine(line);		          	
		
		predict_and_suggest();	
	}
	public void stopDrawingAndErase(){
		snap = null;

		PlaySound.play("sounds\\po.wav");
	
		newLine.removeFromScenegraph();
		
		erase(startPoint);
		operation_status = NONE;	

		predict_and_suggest();	
		
	}



















	//
	// Camera Control
	//

	
	public void rotate(int x, int y){
		Vector3 pos = pickedVehicle.getPosition(jaliceFrame.scene);
		pickedVehicle.setPosition(new Vector3(), jaliceFrame.scene);
		pickedVehicle.rotate(Vector3.Y_AXIS, -(x-prev_x)/100.0, jaliceFrame.scene);//cameraVehicle);
		pickedVehicle.rotate(Vector3.X_AXIS, -(y-prev_y)/100.0, jaliceFrame.scene);//cameraVehicle);
		pickedVehicle.setPosition(pos, jaliceFrame.scene);


		prev_x = x;
		prev_y = y;
	}

	public void slide(int x, int y){
		pickedVehicle.translate(new Vector3((x-prev_x)/100.0,-(y-prev_y)/100.0,0), jaliceFrame.cameraVehicle);


		prev_x = x;
		prev_y = y;
	}


	Transformable dammyVehicle;
	public void orbit(int x, int y){

		jaliceFrame.cameraVehicle.rotate(Vector3.Y_AXIS, (x-prev_x)/100.0, centerBall.jointVehicle);
		
		if (dammyVehicle == null)
			dammyVehicle = new Transformable();
		dammyVehicle.setParent(jaliceFrame.scene);
		dammyVehicle.setPosition(centerBall.position, jaliceFrame.scene);
		dammyVehicle.setOrientation(new Vector3(0,0,1), Vector3.Y_AXIS, (ReferenceFrame) jaliceFrame.cameraVehicle);
		jaliceFrame.cameraVehicle.rotate(Vector3.X_AXIS, (y-prev_y)/100.0, dammyVehicle);


		jaliceFrame.update_light_position();
	
		dammyVehicle.setParent(null);

	
		total_drag += teddy.Vector2.distance(x, y, prev_x, prev_y);
		prev_x = x;
		prev_y = y;
	}
	public void zoom(int x, int y){
		if (zoom_vector == null)
			jaliceFrame.cameraVehicle.translate(new Vector3(0,0,(y-prev_y)/10.0), jaliceFrame.cameraVehicle);
		else
			jaliceFrame.cameraVehicle.translate(Vector3.multiply(zoom_vector, (y-prev_y)/10.0), jaliceFrame.scene);

		total_drag += teddy.Vector2.distance(x, y, prev_x, prev_y);
		prev_x = x;
		prev_y = y;
	}
	double shift_ratio;
	public void shift(int x, int y){
		if (shift_ratio == 0)
			jaliceFrame.cameraVehicle.translate(new Vector3(-(x-prev_x)/100.0,(y-prev_y)/100.0,0), jaliceFrame.cameraVehicle);
		else {
			Vector3 screenPt0 = Chateau.jaliceFrame.screen_coords_to_scene_coords(new Point(prev_x, prev_y));
			Vector3 screenPt1 = Chateau.jaliceFrame.screen_coords_to_scene_coords(new Point(x, y));
			Vector3 vec = Vector3.subtract(screenPt0, screenPt1);
			jaliceFrame.cameraVehicle.translate(Vector3.multiply(vec, shift_ratio), jaliceFrame.scene);
		}
		jaliceFrame.update_light_position();

		total_drag += teddy.Vector2.distance(x, y, prev_x, prev_y);
		prev_x = x;
		prev_y = y;
	}







	public void finish_camera(Point p){
		if (total_drag<10){
			if (centerBall.isVisible())
				hideCenterBall();
			else
				setCenterBall();
				
		}
		else
			hideCenterBall();
	}
	
	public void setCenterBall(){
		if (camera_pick_position != null){
			PlaySound.play("sounds\\click.wav");
			centerBall.setPosition(camera_pick_position);
			centerBall.show();
		}
	}
	public void hideCenterBall(){
		centerBall.hide();
	}
	public boolean centerBall_is_visible(){
		return centerBall.isVisible();
	}
	
	
	Vector3 camera_pick_position;
	Vector3 zoom_vector;
	public void start_camera(Point p){
		camera_pick_position = null;
		shift_ratio = 0;
		zoom_vector = null;
		
		Object pickedStart = pickPlatePlane(p);
		if (pickedStart == null)	
			return;

		Projectable projectable;
		if (pickedStart instanceof Plate){
			Plate plate = (Plate) pickedStart;
			Plane plane = plate.getPlane();
			projectable = new Projectable(plane.base, plane.normal);
		}
		else if (pickedStart instanceof Plane){
			basePlane = (Plane) pickedStart;
			projectable = new Projectable(basePlane.base, basePlane.normal);
		}
		else
			return;

		camera_pick_position = projectable.project(p);

		
		// prepare for shift
		Vector3 cameraPt = Chateau.jaliceFrame.cameraVehicle.getPosition(null);
		Vector3 screenPt = Chateau.jaliceFrame.screen_coords_to_scene_coords(p);
		shift_ratio = Line.distance(camera_pick_position, cameraPt)/ Line.distance(screenPt, cameraPt);

		// prepare for zoom
		zoom_vector = Vector3.subtract(camera_pick_position, jaliceFrame.cameraVehicle.getPosition(null));	//centerBall.position
		if (zoom_vector.getLength() > 0.1)
			zoom_vector.normalize();
		else
			zoom_vector = null;
	}

















	
	
	
	//
	// Event handling
	//


	boolean prediction_switch = false;
	boolean suggestion_switch = true;
	public void keyPressed(java.awt.event.KeyEvent e) {
		//if (e.getKeyCode() == e.VK_SPACE){
		//	flush_candidates();
		//	switchPlane();
		//}
		//else 
		if (e.getKeyCode() == e.VK_P){
			prediction_switch = !prediction_switch;
			if (!prediction_switch)
				flush_candidates();
			else
				makePredictions();
			repaint();
		}
		else if (e.getKeyCode() == e.VK_S){
			suggestion_switch = !suggestion_switch;
			if (!suggestion_switch)
				clearSuggestions();
			else
				makeSuggestions();
			repaint();
		}
		//else walkthrough.keyPressed(e);
		
	}
	public void keyReleased(java.awt.event.KeyEvent e) {
	}
	public void keyTyped(java.awt.event.KeyEvent e) {
	}





    public void mouseClicked( java.awt.event.MouseEvent mouseEvent ) {
    }
    public void mouseEntered( java.awt.event.MouseEvent mouseEvent ) {
    }
    public void mouseExited( java.awt.event.MouseEvent mouseEvent ) {
    }
    public void mouseMoved( java.awt.event.MouseEvent mouseEvent ) {
    }


	Visual pickedVisual;
	Transformable pickedVehicle;

	static int operation_status;
	static final int NONE = 0;
	static final int DRAWING = 1;
	static final int ROTATING = 2;
	static final int SLIDING = 3;

	static final int ORBITING = 4;
	static final int ZOOM = 5;
	static final int SHIFT = 6;
	static final int ZOOM_OR_SHIFT = 7;
	
	static final int SUGGESTION = 8;

    public void mouseReleased( java.awt.event.MouseEvent mouseEvent ) {


		switch(operation_status){
		case DRAWING:
			//finishStroke(mouseEvent);
			finish_drawing(mouseEvent.getPoint(), mouseEvent);
			break;
		case ROTATING:
			if (pickedVisual != null){
            //pickedVisual.setAmbientColor( Color.WHITE );
            //pickedVisual.setDiffuseColor( Color.WHITE );
			}
			break;

		case ZOOM:
		case SHIFT:
		case ORBITING:
		case ZOOM_OR_SHIFT:
			finish_camera(mouseEvent.getPoint());
			break;
			
		case SUGGESTION:
			finishSuggestion(mouseEvent.getPoint());
			break;
		}
		
		operation_status = NONE;
		repaint();

	}
    public void mousePressed( java.awt.event.MouseEvent mouseEvent ) {



		if ((mouseEvent.getModifiers() & mouseEvent.BUTTON3_MASK) != 0 ||
			(mouseEvent.getModifiers() & mouseEvent.SHIFT_MASK) != 0 ){
			start_camera(mouseEvent.getPoint());
			
		    //pickedVisual = pickVisual(mouseEvent.getX(), mouseEvent.getY());
			java.awt.Dimension size = jaliceFrame.awtComponent.getSize();
			if (centerBall_is_visible() || mouseEvent.getX()<40 || mouseEvent.getX()>size.width-40 || mouseEvent.getY()<40 || mouseEvent.getY()>size.height-40)
				operation_status = ORBITING;
			else
				operation_status = ZOOM_OR_SHIFT;
		} 
		else {
			if (pressSuggestion(mouseEvent.getPoint())) {
				dragSuggestion(mouseEvent.getPoint());
				operation_status = SUGGESTION;
			}
			else {
				operation_status = DRAWING;
				start_drawing(mouseEvent.getPoint());
				//add_point(mouseEvent.getPoint());
           	    //pickedVisual = null;
           	}
        }

		prev_x = mouseEvent.getX();
		prev_y = mouseEvent.getY();
		total_drag = 0;
		
		
    }
	int prev_x;
	int prev_y;
	double total_drag;

	public Visual pickVisual(int x, int y){
		edu.cmu.cs.stage3.alice.scenegraph.renderer.PickInfo pickInfo 
			= jaliceFrame.renderTarget.pick( x,y, false, false);//edu.cmu.cs.stage3.alice.scenegraph.renderer.PickStyle.USE_GEOMETRY );

		if (pickInfo == null)
			return null;

        Visual[] visuals = pickInfo.getVisuals();

		if (visuals == null)
			return null;

		if (visuals.length==0)
			return null;

		return visuals[0];

	}

    public void mouseDragged( java.awt.event.MouseEvent mouseEvent ) {

		int x = mouseEvent.getX();
		int y = mouseEvent.getY();
		
		switch(operation_status){
		case DRAWING:
			drawing(mouseEvent.getPoint());
			//add_point(mouseEvent.getPoint());
			//jaliceFrame.repaintOverlay();//paint(renderTarget);
			break;
		case ROTATING:
			rotate(mouseEvent.getX(),mouseEvent.getY());
			break;
		case SLIDING:
			slide(mouseEvent.getX(),mouseEvent.getY());
			break;
		case ORBITING:
			orbit(mouseEvent.getX(),mouseEvent.getY());
			break;
		case ZOOM_OR_SHIFT:
			if (Math.abs(mouseEvent.getX()-prev_x) > Math.abs(mouseEvent.getY()-prev_y)){
				operation_status = SHIFT;
				shift(x,y);
			}
			else{
				operation_status = ZOOM;
				zoom(x,y);
			}
			break;
		case ZOOM:
			zoom(x,y);
			break;
		case SHIFT:
			shift(x,y);
			break;
			
		case SUGGESTION:
			dragSuggestion(mouseEvent.getPoint());
			break;
		}
		
    }


















	public Plane ground;
	public void createGround(){
		ground = new Plane(new Vector3(0,-1,0), Vector3.Y_AXIS);
		ground.setGround();
		ground.addToScenegraph(jaliceFrame.scene);
		
	}
	public CenterBall centerBall;
	public void createCenterBall(){
		centerBall = new CenterBall(new Vector3(0,-1,0));
		centerBall.addToScenegraph(jaliceFrame.scene);
	}

	public void scenegraphCreated(){
		
		helpers = HelperCenter.getHelpers();
		
		init_scene();
		
		//walkthrough = new Walkthrough(this);
	}


	Walkthrough walkthrough;

	// start!
	Chateau(){
	}

	static public JaliceFrame jaliceFrame;
	static public String renderername;
    public static void main( String[] args ) {
		String classname;
		if( args.length>0 ) {
			classname = args[0];
		} else {
			//classname = "edu.cmu.cs.stage3.alice.scenegraph.renderer.dx3renderer.RendererInfo";
			classname = "edu.cmu.cs.stage3.alice.scenegraph.renderer.directx7renderer.RendererInfo";
		}
		renderername = classname;
        try {
            Class cls = Class.forName( classname );
            edu.cmu.cs.stage3.alice.scenegraph.renderer.RendererInfo rendererInfo = (edu.cmu.cs.stage3.alice.scenegraph.renderer.RendererInfo)cls.newInstance();
			Chateau chateau = new Chateau();
            jaliceFrame = new JaliceFrame( rendererInfo, chateau);
        } catch( Throwable t ) {
            t.printStackTrace();
        }
	}
}


