package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;

public class Circle
{
	Vector3 center;
	double radius;
	Circle(Vector3 c, double r){
		center = c;
		radius = r;
	}
	public boolean equivalent(Circle circle){
		return (Joint.equivalent(center, circle.center) && Math.abs(radius-circle.radius)<Chateau.ERROR);
	}
	
	public double getDistance(Vector3 p){
		double r = Line.distance(center, p);
		return Math.abs(r-radius);
	}
	public Vector3 getClosestPoint(Vector3 p){
		Vector3 vec = Vector3.subtract(p, center);
		double r = Line.distance(center, p);
		return Vector3.add(center, Vector3.multiply(vec, radius/r));
	}	
	public Vector3[] crossPoints(InfiniteLine line){
		Vector3 base = line.base;
		Vector3 direction = line.direction;
		
		double[] t = crossPoints_sub(base, direction);
//System.out.println("crossPoints "+t);		
		
		if (t==null)
			return null;
		
		Vector3[] p = new Vector3[t.length];	
		for(int i=0; i<t.length; i++){
			p[i] = Vector3.add(base, Vector3.multiply(direction, t[i]));
		}
		return p;
	}
	public Vector3[] crossPoints(Line line){
		Vector3 base = line.start;
		Vector3 direction = line.getVector3();
		direction.normalize();
		
		double[] t = crossPoints_sub(base, direction);

		if (t==null)
			return null;
		
		Vector p = new Vector();	
		double length = line.getLength();
		for(int i=0; i<t.length; i++){
			if (t[i]>-Chateau.ERROR && t[i]<length+Chateau.ERROR)
				p.addElement(Vector3.add(base, Vector3.multiply(direction, t[i])));
		}
		Vector3[] q = new Vector3[p.size()];
		for(int i=0; i<p.size(); i++)
			q[i] = (Vector3) p.elementAt(i);
		return q;
	}



	public double[] crossPoints_sub(Vector3 base, Vector3 direction){
		
		Vector3 center_to_base = Vector3.subtract(base, center);

		if (Math.abs(Vector3.crossProduct(center_to_base, direction).getLength() - radius) < Chateau.ERROR){
			//System.out.println("tangent");
			double t = Vector3.dotProduct(center_to_base, direction);
			double[] ts = {t};
			return ts;
			//Vector3 v = Vector3.add(base, Vector3.multiply(direction, -t));
			//Vector3[] result = {v};
			//return result;
		
		}
					
		// | base + t * direction - center | == radius
		// (dx2+dy2+dz2)*t2 + 2(dx(bx-cx)+dy(by-cy)+dz(bz-cz))*t + (bx-cx)2+(by-cy)2+(bz-cz)2-r2 == 0
		double a =  direction.x * direction.x + 
					direction.y * direction.y + 
					direction.z * direction.z ;
		double b = 2*(direction.x * center_to_base.x +
					direction.y * center_to_base.y +
					direction.z * center_to_base.z );
		double c = center_to_base.x * center_to_base.x +
					center_to_base.y * center_to_base.y +
					center_to_base.z * center_to_base.z - radius*radius;
		// a*x*x + b*x + c = 0
		if (b*b-4*a*c < 0)
			return null;
			
		double d = Math.sqrt(b*b-4*a*c);
		double t0 = (-b+d)/2/a;
		double t1 = (-b-d)/2/a;
		
		double[] t = {t0, t1};
		return t;
		
		//		Vector3 v0 = Vector3.add(base, Vector3.multiply(direction, t0));
		//		Vector3 v1 = Vector3.add(base, Vector3.multiply(direction, t1));
		
		//		Vector3[] result = {v0, v1};
		//		return result;
	}
}