package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class ColumnHelper extends PlaneHelper
{
	public Suggestion makeSuggestion(Chateau chateau){
	
		Vector baseLines = Util.duplicateVector(chateau.currentLines);
		if (baseLines.size() == 0 || !all_connected(baseLines))
			return null;
		
		/*
		Plane basePlane = find_common_plane(baseLines);
		
		if (basePlane==null)
			return null;
		if (basePlane != chateau.ground){
			Plate basePlate = null;
			for(int i=0; i<chateau.plates.size(); i++){
				Plate plate = (Plate)chateau.plates.elementAt(i);
				if (basePlane.contains(plate) && contain_all_lines(plate, baseLines)){
					basePlate = plate;
					break;
				}
			}
			if (basePlate == null)
				return null;
		}
		*/
		
		Plane basePlane = null;
		if (contain_all_lines(chateau.ground, baseLines))
			basePlane = chateau.ground;
		else {
			Plate basePlate = null;
			for(int i=0; i<chateau.plates.size(); i++){
				Plate plate = (Plate)chateau.plates.elementAt(i);
				if (contain_all_lines(plate, baseLines)){
					basePlate = plate;
					break;
				}
			}
			if (basePlate == null)
				return null;
			basePlane = basePlate.getPlane();
		}
		
		
		Vector sorted_plates = get_sorted_list_of_possible_plates(basePlane, chateau);
		if (sorted_plates == null)
			return null;
			
			
		
		Vector topLines = null;
		for(int i=0; i<sorted_plates.size(); i++){
			Plate plate = (Plate) sorted_plates.elementAt(i);
			Vector lines = getTopLines(plate, baseLines, basePlane);
			if (lines != null){
				topLines = lines;
				break;
			}
		}
		if (topLines == null)
			return null;
			


		// let us make a suggestion!!!
			
		World originalWorld = 	new World(chateau);

		
		for(int i=0; i<baseLines.size(); i++){
		
			Line a = (Line) baseLines.elementAt(i);

			Line b = (Line) topLines.elementAt(i);
			Joint j0 = chateau.findJoint(b.start);
			Joint j1 = chateau.findJoint(b.end);
			b = chateau.addLine(new Line(j0, j1));
			
			Line line0;
			Line line1;
			if (a.smaller_than_90_degree(b)){
				line0 = chateau.addLine(new Line(a.startJoint, b.startJoint));
				line1 = chateau.addLine(new Line(a.endJoint, b.endJoint));
			}
			else {
				line0 = chateau.addLine(new Line(a.startJoint, b.endJoint));
				line1 = chateau.addLine(new Line(a.endJoint, b.startJoint));
			}
			
			chateau.addCurrentLine(line0);
			chateau.addCurrentLine(line1);
			
			chateau.addPlate(new Plate(a, line0, b, line1));
		}
		
		Suggestion suggestion = new Suggestion(chateau, "Column");
		chateau.restoreWorld(originalWorld);
		return suggestion;

	}
	
	public Vector getTopLines(Plate plate, Vector baseLines, Plane basePlane){
		
		if (!basePlane.parallel(plate.getPlane())) 
			return null;
		if (basePlane.contains(plate))
			return null;
		
		double t = Vector3.dotProduct(basePlane.normal, Vector3.subtract(plate.base(), basePlane.base));
		Vector3 shiftVector = Vector3.multiply(basePlane.normal, t);
		
		Vector topLines = new Vector();
		for(int i=0; i<baseLines.size(); i++){
			Line line = (Line) baseLines.elementAt(i);
			Line topLine = new Line(Vector3.add(line.start, shiftVector), Vector3.add(line.end, shiftVector));
			if (!plate.on_plate(topLine))
				return null;
			topLines.addElement(topLine);
		}
		return topLines;
		
	
	}
	
	public Vector get_sorted_list_of_possible_plates(Plane basePlane, Chateau chateau){
		Vector sortedPlates = new Vector();
		
		Vector3 cameraPosition = chateau.jaliceFrame.cameraVehicle.getPosition(null);
		double sign = Vector3.dotProduct(basePlane.normal, Vector3.subtract(cameraPosition, basePlane.base));
		
		for(int i=0; i<chateau.plates.size(); i++){
			Plate plate = (Plate) chateau.plates.elementAt(i);
			Plane plane = plate.getPlane();
			if (!basePlane.contains(plate) && basePlane.parallel(plane)){
				double d = ditance_from_basePlane(basePlane, plane)*sign;
				if (d>0){
					insert_in_sorted_list(sortedPlates, new SortedPlate(d,plate));
				}
			}
		}
		if (sortedPlates.size() == 0)
			return null;
			
		Vector plates = new Vector();
		for(int i=0; i<sortedPlates.size(); i++){
			SortedPlate sp = (SortedPlate) sortedPlates.elementAt(i);
			plates.addElement(sp.plate);
		}
		return plates;
			
	}
	public double ditance_from_basePlane(Plane basePlane, Plane plane){
		return Vector3.dotProduct(basePlane.normal, Vector3.subtract(plane.base, basePlane.base));		
	}
	
	class SortedPlate {
		double d;
		Plate plate;
		SortedPlate(double d, Plate plate){
			this.d = d;
			this.plate = plate;
		}
	}
	public void insert_in_sorted_list(Vector list, SortedPlate element){
		for(int i=0; i<list.size(); i++){
			SortedPlate sp = (SortedPlate) list.elementAt(i);
			if (element.d < sp.d){
				list.insertElementAt(element, i);
				return;
			}
		}
		list.addElement(element);
	}
	public boolean contain_all_lines(Plate plate, Vector lines){
		for(int j=0; j<lines.size(); j++){
			Line line = (Line) lines.elementAt(j);
			if (!plate.on_plate(line))
				return false;
		}
		return true;
	}	
	public boolean contain_all_lines(Plane plane, Vector lines){
		for(int j=0; j<lines.size(); j++){
			Line line = (Line) lines.elementAt(j);
			if (!plane.contains(line))
				return false;
		}
		return true;
	}	
}
