package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class CornerCutHelper extends Helper
{
	int size;
	public Suggestion makeSuggestion(Chateau chateau){
		if (chateau.currentLines.size() <=1)
			return null;


		Vector loopLines;
		Vector loopJoints;
			
		Object[] loop = get_loop(chateau.currentLines);
		if (loop != null){
			loopLines = (Vector) loop[0];
			loopJoints = (Vector) loop[1];
			size = loopLines.size();
		}
		else {
			Object[] chain = get_chain(chateau.currentLines);
			if (chain == null)
				return null;
			loopLines = (Vector) chain[0];
			loopJoints = (Vector) chain[1];
		}
			
		size = loopJoints.size();

		Vector inactiveLines = new Vector();
		for(int i=0; i<chateau.lines.size(); i++){
			Line line = (Line) chateau.lines.elementAt(i);
			if (!line.active)
				inactiveLines.addElement(line);
		}

		
		
		//
		Joint[] sideJoints = new Joint[size];
		sideJoints[0] = (Joint) loopJoints.elementAt(0);
		
		Vector possibilities = find_possibilities(inactiveLines, sideJoints[0]);
		if (possibilities.size() == 0)
			return null;
		
		
		for(int i=1; i<size; i++){
			sideJoints[i] = (Joint) loopJoints.elementAt(i);
			possibilities = narrow_possibilities(possibilities, sideJoints[i]);
			if (possibilities.size() == 0)
				return null;
		}
		if (possibilities.size() != 1)
			return null;
		
		
		Possibility possibility = (Possibility) possibilities.elementAt(0);
		Joint cornerJoint = possibility.cornerJoint;
		Line[] sideLines = possibility.getSideLines();
		
		

		Plate[] sidePlates = new Plate[size];
		Line line0 = sideLines[size-1];
		for(int i=0; i<size; i++){
			Line line1 = sideLines[i];
			sidePlates[i] = find_common_plate(line0, line1);
			if (sidePlates[i] == null)		return null;
			line0 = line1;
		}

		








		
		// let's make a suggestion!!	
		World originalWorld = 	new World(chateau);


		if (loopLines.size() < size){
			Line line = chateau.addLine(new Line(
									(Joint)loopJoints.elementAt(0), 
									(Joint)loopJoints.elementAt(size-1) ));
			loopLines.insertElementAt(line, 0);
			chateau.addCurrentLine(line);
		}


		Line[] cornerLines = new Line[size];
		for(int i=0; i<size; i++){
			cornerLines[i] = (Line) loopLines.elementAt(i);
		}


		Line[] newSideLines = new Line[size];
		for(int i=0; i<size; i++){
			Joint joint = sideLines[i].get_the_other_joint(cornerJoint);
			if (sideJoints[i] != joint)
				newSideLines[i] = chateau.addLine(new Line(sideJoints[i], joint));
		}

		for(int i=0; i<size; i++){
			Vector plateLines = Util.duplicateVector(sidePlates[i].lines);
			plateLines.removeElement(sideLines[Util.mod(i-1, size)]);
			plateLines.removeElement(sideLines[i]);
			if (newSideLines[Util.mod(i-1, size)] != null)
				plateLines.addElement(newSideLines[Util.mod(i-1, size)]);
			if (newSideLines[i] != null)
				plateLines.addElement(newSideLines[i]);
			plateLines.addElement(cornerLines[i]);
			if (plateLines.size() >=3)
				chateau.addPlate(new Plate(plateLines));
		}

		chateau.addPlate(new Plate(loopLines));
		

		for(int i=0; i<size; i++){
			chateau.deletePlate(sidePlates[i]);
		}

		for(int i=0; i<size; i++){
			chateau.deleteLine(sideLines[i]);
		}

		

		Suggestion suggestion = new Suggestion(chateau, "CornerCut");
		chateau.restoreWorld(originalWorld);
		return suggestion;
		
	}
	public Line find_touched_line(Vector lines, Joint joint){
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (line.on_line_segment(joint.position) && ! line.contains(joint))
				return line;
		}
		return null;
	}
	
	public Plate find_common_plate(Line line0, Line line1){
		for(int i=0; i<line0.plates.size(); i++){
			Plate plate = (Plate) line0.plates.elementAt(i);
			if (line1.plates.contains(plate))
				return plate;
		}
		return null;
	}





	class Possibility {	
		Joint cornerJoint;
		Vector sideLines;
		Possibility(Joint j, Line line){
			cornerJoint = j;
			sideLines = new Vector();
			sideLines.addElement(line);
		}
		public void addLine(Line line){
			sideLines.addElement(line);
		}
		public Line[] getSideLines(){
			Line[] lines = new Line[size];
			for(int i=0; i<sideLines.size(); i++){
				lines[i] = (Line) sideLines.elementAt(i);
			}
			return lines;
		}
	}
	public Vector find_possibilities(Vector lines, Joint joint){
		Vector possibilities = new Vector();
		
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (line.on_line_segment(joint.position)){
				if (line.contains(joint)){
					possibilities.addElement(new Possibility(line.get_the_other_joint(joint), line));
				}
				else{
					possibilities.addElement(new Possibility(line.startJoint, line));
					possibilities.addElement(new Possibility(line.endJoint, line));
				}
			}
		}
		return possibilities;
	
	}
	public Vector narrow_possibilities(Vector possibilities, Joint joint){
		Vector result = new Vector();
		for(int i=0; i<possibilities.size(); i++){
			Possibility pos =(Possibility) possibilities.elementAt(i);
			for(int j=0; j<pos.cornerJoint.lines.size(); j++){
				Line line = (Line) pos.cornerJoint.lines.elementAt(j);
				if (!pos.sideLines.contains(line) && line.on_line_segment(joint.position)){
					pos.addLine(line);
					result.addElement(pos);
					break;
				}
			}
		}
		return result;
	}
	
}
