package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;

// prototype for various ***Helper classes

public abstract class Helper
{
	Helper(){}

	// you will implement this method for your ***Helper 
	public abstract Suggestion makeSuggestion(Chateau chateau);

	
	


	// frequently-used common methods
	
	public Plane find_common_plane(Vector lines){
		Plane commonPlane = null;
		if (lines.size() <= 1)
			return null;
		
		Line baseLine = (Line) lines.elementAt(0);
		for(int i=1; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (!baseLine.on_line(line)){
				Plane plane = baseLine.get_common_plane(line);
				if (plane == null)
					return null;
				else if ( commonPlane == null){
					commonPlane = plane;
				}
				else if (!commonPlane.equivalent(plane))
					return null;
			}
		}
		return commonPlane;
	}
	
	



	// CornerCut
	public Object[] get_loop(Vector lines){
		Vector visited = new Vector();
		Vector joints = new Vector();
		
		Line baseLine = (Line) lines.elementAt(0);
		Line prevLine = baseLine;
		Joint prevJoint = baseLine.endJoint;
		visited.addElement(baseLine);
		joints.addElement(prevJoint);
		while(true){
			prevLine = findConnectedLine(prevJoint, prevLine, lines);
			if (prevLine == null)
				return null;
			if (prevLine == baseLine)
				break;
			if (visited.contains(prevLine))
				return null;
			prevJoint = prevLine.get_the_other_joint(prevJoint);
			visited.addElement(prevLine);
			joints.addElement(prevJoint);
		}	
		if (visited.size() != lines.size())
			return null;
		
		Object[] result = {visited, joints};
		return result;
	}


	// used by Extrusion, Pyramid	
	public boolean is_loop(Vector lines){
		Vector visited = new Vector();
		
		Line baseLine = (Line) lines.elementAt(0);
		Line prevLine = baseLine;
		Joint prevJoint = baseLine.endJoint;
		visited.addElement(baseLine);
		while(true){
			prevLine = findConnectedLine(prevJoint, prevLine, lines);
			if (prevLine == null)
				return false;
			if (prevLine == baseLine)
				break;
			if (visited.contains(prevLine))
				return false;
			prevJoint = prevLine.get_the_other_joint(prevJoint);
			visited.addElement(prevLine);
		}	
		
		return (visited.size() == lines.size());
	}
	public Line findConnectedLine(Joint joint, Line prevLine, Vector lines){
		
		Line nextLine = null;
		for(int i=0; i<joint.lines.size(); i++){
			Line line = (Line) joint.lines.elementAt(i);
			if (line != prevLine && lines.contains(line)){
				if (nextLine == null)
					nextLine = line;
				else
					return null;
			}
		}
		return nextLine;
	}	
	











	
	// used by Extrusion, Pyramid	
	public boolean all_connected(Vector lines){
		if (lines.size() == 0)
			return false;
			
		boolean[] visited = new boolean[lines.size()];
		Line line = (Line) lines.elementAt(0);
		visited[0] = true;
		all_connected_sub(line, line.startJoint, lines, visited);
		all_connected_sub(line, line.endJoint, lines, visited);
		
		int count=0;
		for(int i=0; i<visited.length; i++){
			if (visited[i])
				count++;
		}
		return (count == visited.length);
	}
	public void all_connected_sub(Line baseLine, Joint joint, Vector lines, boolean[] visited){
		for(int i=0; i<joint.lines.size(); i++){
			Line line =  (Line) joint.lines.elementAt(i);
			int index = lines.indexOf(line);
			if (index != -1 && !visited[index]){
				visited[index] = true;
				all_connected_sub(line, line.get_the_other_joint(joint), lines, visited);
			}
		}
	}
	
	// from FrameHelper
	public double total_length(Vector lines){
		double l = 0;
		for(int i=0; i<lines.size(); i++){
			Line line =  (Line) lines.elementAt(i);
			l += line.getLength();
		}
		return l;
	}


	//chamfer, cornercut
	public Object[] get_chain(Vector lines){
		if (lines.size() == 0)
			return null;
			
		Line baseLine = (Line) lines.elementAt(0);
			
		Object[] chain0 = get_chain_sub(baseLine, baseLine.startJoint, lines);
		if (chain0 == null)
			return null;
		Object[] chain1 = get_chain_sub(baseLine, baseLine.endJoint, lines);
		if (chain1 == null)
			return null;
		
		Vector result = Util.reverseVector((Vector)chain0[0]);
		result.addElement(baseLine);
		result = Util.connectVector(result, (Vector)chain1[0]);
		
		if (result.size() != lines.size())
			return null;
			
		Vector joints = Util.reverseVector((Vector)chain0[1]);
		joints.addElement(baseLine.startJoint);
		joints.addElement(baseLine.endJoint);
		joints = Util.connectVector(joints, (Vector)chain1[1]);
		
		Object[] objects = {result, joints};
		return objects;
	}
	public Object[] get_chain_sub(Line baseLine, Joint prevJoint, Vector lines){
		Vector visited = new Vector();
		Vector joints = new Vector();
		
		Line prevLine = baseLine;
		while(true){
			prevLine = findConnectedLine(prevJoint, prevLine, lines);
			if (prevLine == null)
				break;
			if (visited.contains(prevLine))
				return null;
			prevJoint = prevLine.get_the_other_joint(prevJoint);
			visited.addElement(prevLine);
			joints.addElement(prevJoint);
		}	
		
		Object[] objects = {visited, joints};
		return objects;
	}
	//chamfer
	public Plate find_common_plate(Vector lines, Chateau chateau){
		if (lines.size() <= 0)
			return null;
		
		Line baseLine = (Line) lines.elementAt(0);

		Vector plates = get_plates_containing(baseLine, chateau);
		if (plates.size() ==0)
			return null;

		for(int i=1; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			Vector updatedPlates = new Vector();
			for(int j=0; j<plates.size(); j++){
				Plate plate = (Plate) plates.elementAt(j);
				if (plate.on_plate(line))
					updatedPlates.addElement(plate);
			}
			if (updatedPlates.size() == 0)
				return null;
			plates = updatedPlates;
		}
		return (Plate) plates.elementAt(0);
	}
	public Vector get_plates_containing(Line line, Chateau chateau){
		Vector plates = new Vector();
		for(int i=0; i<chateau.plates.size(); i++){
			Plate plate = (Plate) chateau.plates.elementAt(i);
			if (plate.on_plate(line)){
				plates.addElement(plate);
				break;
			}
		}
		return plates;
	}	
}
