package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class Joint extends ChateauElement
{
	public int index;

	Vector3 position;
	
	public Vector lines = new Vector();	

	final double RADIUS = 0.06;//0.08;
	Color COLOR = Def.JOINT_COLOR;	//new Color(0.4,0.4,0.4,1);//Color.WHITE;//new Color(0.2,0.2,0.2,1);//Color.BLACK;//white
	
	Joint(Vector3 pos){
		this.position = pos;
	}
	Joint(){
	}
	
	public void refresh_network(Hashtable hash){
		Vector revised_lines = new Vector();
		for(int i=0; i<lines.size(); i++){
			Line original_line = (Line) lines.elementAt(i);
			Line line = (Line) hash.get(original_line);
			revised_lines.addElement(line);
		}
		lines = revised_lines;
	}
	public Joint duplicate(){
		Joint joint = new Joint();
		joint.position = position;
		joint.lines = lines;
		
		joint.jointVehicle = jointVehicle;
		joint.jointVisual = jointVisual;
		
		return joint;
	}
	
	
	public Line getCommonLine(Joint joint){
		for(int i=0; i<lines.size(); i++){
			Line line = (Line)lines.elementAt(i);
			if (joint.lines.contains(line))
				return line;
		}
		return null;
		
	}


	static public boolean equivalent(Vector3 p0, Vector3 p1){
		return (Vector3.subtract(p1, p0).getLength() < Chateau.ERROR);
	}
	public boolean equivalent(Vector3 p1){
		return equivalent(position, p1);
	}
	public boolean equivalent(Joint j){
		return equivalent(position, j.position);
	}


	public double distance(Joint joint){
		return (Vector3.subtract(position, joint.position).getLength());
	}




	public boolean active;
	public void activate(){
		active = true;
		//jointVisual.setDiffuseColor(Color.RED);
		frontFacingAppearance.setDiffuseColor(Color.RED);

	}
	public void disactivate(){
		active = false;
		//jointVisual.setDiffuseColor(new Color(1,1,1));
		frontFacingAppearance.setDiffuseColor(new Color(1,1,1));
	}
	
	// centerBall
	public void hide(){
		jointVisual.setParent(null);
	}
	public void show(){
		jointVisual.setParent(jointVehicle);
	}
	public boolean isVisible(){
		return (jointVisual.getParent() != null);
	}
	
	
	public void delete(){
		jointVehicle.setParent(null);
	}
	public void deleteLine(Line line){
		lines.removeElement(line);
		
		if (lines.size() == 0){
			Chateau.joints.removeElement(this);
			delete();
		}
	}

	Transformable jointVehicle;
	Visual jointVisual;

	public void removeFromScenegraph(){
		if (jointVehicle != null){
			jointVehicle.setParent(null);
		}
	}
	Scene scene;
	public void addToScenegraph(Scene parent){
		scene = parent;
		if (jointVehicle == null){
			createJointVisual(position);

			jointVehicle = new Transformable();
       		jointVisual.setParent(jointVehicle);
       	}
		jointVehicle.setParent( parent );
       	jointVehicle.setPosition(position, parent);
	}
	public Transformable getVehicle(){
		return jointVehicle;
	}
	


	Appearance frontFacingAppearance = null;
	public void createJointVisual(Vector3 pos){
	
		jointVisual = new Visual();

		//jointVisual.setFillingStyle(FillingStyle.SOLID);
		//jointVisual.setShadingStyle(ShadingStyle.SMOOTH);
		frontFacingAppearance = new Appearance();
		frontFacingAppearance.setFillingStyle(FillingStyle.SOLID);
		frontFacingAppearance.setShadingStyle(ShadingStyle.SMOOTH);
		jointVisual.setFrontFacingAppearance(frontFacingAppearance);
		
		setGeometry(RADIUS);
	}

	
	public void setGeometry(double radius){
		
		if (jointITA == null)
			jointITA = createITA(radius);

        jointVisual.setGeometry( jointITA );
		
		//jointVisual.setDiffuseColor(COLOR);
        frontFacingAppearance.setDiffuseColor(COLOR);
	}
	
	
	
	public static IndexedTriangleArray jointITA;
	public IndexedTriangleArray createITA(double radius){
	
		jointITA = new IndexedTriangleArray();

		Vector3 center = new Vector3(0,0,0);
		Vector3 nv;
		Vertex[] jointVertices = new Vertex[2+3*8];
		int n=0;
		for(int y=0; y<5; y++){
			double yangle = Math.PI*2*y/8;
			double yradius = radius * Math.cos(yangle);
			double xradius = radius * Math.sin(yangle);
			if (y==0 || y==4){
				int i = n++;
				jointVertices[i] = new Vertex(Vertex.POSITION_XYZ|Vertex.NORMAL_IJK|Vertex.TEXTURE_COORDINATE_UV0);
				Vector3 v =	new Vector3(0, yradius, 0);
				jointVertices[i].setXYZ(v);
				nv = Vector3.subtract(v, center); nv.normalize();
				jointVertices[i].setIJK(nv);
				jointVertices[i].setU(0);
				jointVertices[i].setV(0);
			}
			else for(int x=0; x<8; x++){
				double xangle = Math.PI*2*x/8;
				int i = n++;
				jointVertices[i] = new Vertex(Vertex.POSITION_XYZ|Vertex.NORMAL_IJK|Vertex.TEXTURE_COORDINATE_UV0);
				Vector3 v =	new Vector3(xradius * Math.cos(xangle),
							yradius,
							xradius * Math.sin(xangle));
				jointVertices[i].setXYZ(v);
				nv = Vector3.subtract(v, center); nv.normalize();
				jointVertices[i].setIJK(nv);
				jointVertices[i].setU(0);
				jointVertices[i].setV(0);
			}
		}

		
		
		
		
		
		int[] jointIndices = new int[2*8*6 + 2*8*3];

		n=0;
		for(int x=0; x<8; x++){
			jointIndices[n++] = 17+x;
			jointIndices[n++] = 17+Util.mod(x+1,8);
			jointIndices[n++] = 25;
					
			jointIndices[n++] = 1+x;
			jointIndices[n++] = 0;
			jointIndices[n++] = 1+Util.mod(x+1,8);
		}
			
		for(int y=0; y<2; y++)
			for(int x=0; x<8; x++){
				jointIndices[n++] = 1+y*8+x;
				jointIndices[n++] = 1+y*8+Util.mod(x+1,8);
				jointIndices[n++] = 1+y*8+x+8;
					
				jointIndices[n++] = 1+y*8+x+8;
				jointIndices[n++] = 1+y*8+Util.mod(x+1,8);
				jointIndices[n++] = 1+y*8+Util.mod(x+1,8)+8;
		}
		
		

		jointITA.setVertices(jointVertices);
		jointITA.setIndices(jointIndices);

		return jointITA;
	}

}
