package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;

public class Line extends ChateauElement
{
	public int index;
	
	Vector3 start;
	Vector3 end;
	
	Joint startJoint;
	Joint endJoint;

	Vector plates = new Vector();
	
	public final Color BASE_COLOR = Def.LINE_BASE_COLOR;	// new Color(0.3, 0.3, 0.3);//0.2,0.2,0.2);
	public final Color HIGHLIGHTED_COLOR = Def.LINE_HIGHLIGHTED_COLOR;	//new Color(1, 0.49, 0.17);//0.9,0,0,1);
	public final Color CANDIDATE_COLOR = Def.LINE_CANDIDATE_COLOR;	//new Color(0.92, 0.58,  1);//0.4,0.0,0.9);
	
	
	public void refresh_network(Hashtable hash){
	
		

		Vector revised_plates = new Vector();
		for(int i=0; i<plates.size(); i++){
			Plate original_plate = (Plate) plates.elementAt(i);
			Plate plate = (Plate) hash.get(original_plate);
			revised_plates.addElement(plate);
		}
		plates = revised_plates;

		startJoint = (Joint) hash.get(startJoint);
		endJoint = (Joint) hash.get(endJoint);
		
	}
	public Line duplicate(){
		Line line = new Line();
		line.start = start;
		line.end = end;
		line.startJoint = startJoint;
		line.endJoint = endJoint;
		
		line.plates = plates;
		
		line.lineVehicle = lineVehicle;
		line.lineVisual = lineVisual;
		line.lineAppearance = lineAppearance;
		
		return line;
	}

	
	
	
	Line(){};
	Line(Plane parent, Vector3 start, Vector3 end){

		this.start = start;
		this.end = end;
		
	}
	Line(Vector3 start, Vector3 end){
		this.start = start;
		this.end = end;
		
	}
	
	
	public boolean equivalent(Line line){
		if (Joint.equivalent(start, line.start) && Joint.equivalent(end, line.end) )
			return true;
		if (Joint.equivalent(start, line.end) && Joint.equivalent(end, line.start) )
			return true;
		return false;
	}
	public boolean equalLength(Line line){
		if (Math.abs(getLength() - line.getLength()) < Chateau.ERROR)
			return true;
		return false;
	}
	public Plane get_common_plane(Line line){

	
		// ensure that they are not on a same line..
		Vector3 vec0 = this.getVector3();
		Vector3 vec1;
		if (this.on_line(line.startJoint.position))
			vec1 = Vector3.subtract(line.endJoint.position, this.startJoint.position);
		else
			vec1 = Vector3.subtract(line.startJoint.position, this.startJoint.position);
		Vector3 base = this.startJoint.position;
		Vector3 normal = Vector3.crossProduct(vec0, vec1);
		normal.normalize();
		Plane plane = new Plane(base, normal);
		
		if (plane.contains(line))
			return plane;
		else 
			return null;
	}	
		
	public boolean on_line_segment(Vector3 p){
		Vector3 vec0 = getVector3();
		Vector3 vec1 = Vector3.subtract(p, start);
		double d = Vector3.crossProduct(vec0, vec1).getLength();
		if (d>Chateau.ERROR) return false;
		
		if (Vector3.dotProduct(vec0, vec1) < -Chateau.ERROR)
			return false;
		Vector3 vec2 = Vector3.subtract(p, end);
		if (Vector3.dotProduct(vec0, vec2) > Chateau.ERROR)
			return false;
			

		return true;
		
		
	}

	public double getLength(){
		return getVector3().getLength();
	}

	public boolean on_line(Line line){
		return (on_line(line.start) && on_line(line.end));
	}	
	public boolean on_line(Vector3 p){
		Vector3 vec0 = getVector3();
		Vector3 vec1 = Vector3.subtract(start, p);
		double d = Vector3.crossProduct(vec0, vec1).getLength();
		return (d<Chateau.ERROR);
	}
	public boolean connected(Line line){
		return (startJoint == line.startJoint || startJoint == line.endJoint ||
				endJoint == line.startJoint || endJoint == line.endJoint);
	}
	public double getDistance(Vector3 p){
		Vector3 vec = getVector3();
		Vector3 startToP = Vector3.subtract(p, start);
		if (Vector3.dotProduct(vec, startToP)<0)
			return distance(start, p);
		Vector3 endToP = Vector3.subtract(p, end);
		if (Vector3.dotProduct(vec, endToP)>0)
			return distance(end, p);
	  	
	  	return getDistance_as_infinite(p);
	}
	public double getDistance_as_infinite(Vector3 p){
		Vector3 vec = getVector3();
		Vector3 startToP = Vector3.subtract(p, start);

		double bunbo = vec.getLength();
		double bunshi = Vector3.crossProduct(vec, startToP).getLength();

		if ( bunbo == 0 ) {
			System.out.println("Error in Line.getDistance()");
	       	return 0;
	    }
	  	else {
	       return Math.abs(bunshi / bunbo);
	  	}
			
	}
	public static double distance(Vector3 a, Vector3 b){
		return (Vector3.subtract(a,b).getLength());
	}
	public Vector3 midPoint(){
		return Vector3.add(start, Vector3.multiply(getVector3(), 0.5));
	}
	static public Vector3 midPoint(Vector3 p, Vector3 q){
		return Vector3.multiply(Vector3.add(p,q), 0.5);
	}
	public Vector3 getFoot(Vector3 p){
		Vector3 vec = getVector3();
		vec.normalize();
		double t = Vector3.dotProduct(Vector3.subtract(p, start), vec);
		return Vector3.add(start, Vector3.multiply(vec, t));
	}
	
	public Vector3 getClosestPoint(Vector3 p){
		Vector3 vec = getVector3();
		double t = Vector3.dotProduct(Vector3.subtract(p, start), vec)/Vector3.dotProduct(vec, vec);
		return Vector3.add(start, Vector3.multiply(vec, t));
	}	

	public boolean onSamePlane(Line b){
		Line a = this;

		// same plane test
		if (a.on_line(b.start) || a.on_line(b.end) ||
			b.on_line(a.start) || b.on_line(a.end))
				return true;
				
		Vector3 start_to_start = Vector3.subtract(a.start, b.start);
		start_to_start.normalize();
		Vector3 normal = Vector3.crossProduct(start_to_start, a.getNormal());
		if (Math.abs(Vector3.dotProduct(normal, b.getNormal())) > Chateau.ERROR)
			return false;
		return true;
	}
	// ƂČ_邩B
	/*
	public boolean cross(Line line){
	  if (!onSamePlane(line))
			return false;
	
	  if (parallel(line))
	  	return false;

	
	  // a * x + b * y + c = 0
	  double a0, b0, c0,
        	 a1, b1, c1; 

	  a0 = start.y - end.y; 
	  b0 = end.x - start.x;
	  c0 = end.y * start.x - end.x * start.y;

	  a1 = line.start.y - line.end.y; 
	  b1 = line.end.x - line.start.x;
	  c1 = line.end.y * line.start.x - line.end.x * line.start.y;

	  if (((a0*line.start.x+b0*line.start.y+c0)*(a0*line.end.x+b0*line.end.y+c0)<= Chateau.ERROR) &&
	      ((a1*     start.x+b1*     start.y+c1)*(a1*     end.x+b1*     end.y+c1)<= Chateau.ERROR))
		return true;
	  else
		return false;
	}
		*/

	// ƂČ_Ƃ߂B
	public Vector3 getCrossPoint(Line line){
	  if (!onSamePlane(line) || parallel(line))
	  	return null;


		// d = |u0+s*u - v0+t*v|;
		// dd/ds = s*u*u - t*v*u + uv*u = 0;
		// dd/dt = s*u*v - t*v*v + uv*v = 0;
		
		Vector3 u = getVector3();
		Vector3 v = line.getVector3();
		Vector3 uv = Vector3.subtract(start, line.start);
		
		// a0*s + b0*t = c0;
		// a1*s + b1*t = c1;
		
		double a0 = Vector3.dotProduct(u,u);
		double b0 = -Vector3.dotProduct(v,u);
		double c0 = -Vector3.dotProduct(uv,u);
		
		double a1 = Vector3.dotProduct(u,v);
		double b1 = -Vector3.dotProduct(v,v);
		double c1 = -Vector3.dotProduct(uv,v);
		
		double d = a0*b1-a1*b0;
		
		double s = (b1*c0-b0*c1)/d;
		double t = (-a1*c0+a0*c1)/d;
		
		if (s<-Chateau.ERROR || s>1+Chateau.ERROR ||
			t<-Chateau.ERROR || t>1+Chateau.ERROR )
			return null;
		
		
		return Vector3.add(start, Vector3.multiply(u, s));
	}

	public boolean smaller_than_90_degree(Line line){
		return (Vector3.dotProduct(getVector3(), line.getVector3()) > 0);
	}
	public boolean parallel(Line line){
		return (Vector3.crossProduct(getVector3(), line.getVector3()).getLength() < Chateau.ERROR);
	}
	public boolean perpendicular(Line line){
		return (Math.abs(Vector3.dotProduct(getVector3(), line.getVector3()))< Chateau.ERROR);
	}
	public static boolean parallel(Vector3 v0, Vector3 v1){
		v0 = Vector3.normalizeV(v0);
		v1 = Vector3.normalizeV(v1);
		return (Vector3.crossProduct(v0, v1).getLength() < Chateau.ERROR);
	}
	public double parallel_distance(Line line){
		Vector3 normal = getNormal();
		Vector3 start_to_start = Vector3.subtract(line.start, start);
		return Vector3.crossProduct(start_to_start, normal).getLength();
	}

	public Vector gatherConnectedLines(){
		return gatherConnectedLines(null);
	}	
	public Vector gatherConnectedLines(Vector base){
		Vector result = new Vector();
		result.addElement(this);
		
		for(int i=0; i<startJoint.lines.size(); i++){
			Line line = (Line) startJoint.lines.elementAt(i);
			if (line != this && !result.contains(line) && (base == null || base.contains(line)))
				line.gatherConnectedLines(startJoint, result, base);
		}
		for(int i=0; i<endJoint.lines.size(); i++){
			Line line = (Line) endJoint.lines.elementAt(i);
			if (line != this && !result.contains(line)  && (base == null || base.contains(line)))
				line.gatherConnectedLines(endJoint, result, base);
		}
		return result;
	
	}
	public void gatherConnectedLines(Joint base_joint, Vector result, Vector base){
		result.addElement(this);
		Joint joint = get_the_other_joint(base_joint);
		for(int i=0; i<joint.lines.size(); i++){
			Line line = (Line) joint.lines.elementAt(i);
			if (line != this && !result.contains(line) && (base == null || base.contains(line)))
				line.gatherConnectedLines(joint, result, base);
		}
		
	}
	
	
	
	public Joint get_the_other_joint(Joint joint){
		if (joint == startJoint)
			return endJoint;
		else if (joint == endJoint)
			return startJoint;
		
		System.out.println("error at Line.get_the_other_joint()");
		return startJoint;
	}
	public Joint get_common_joint(Line line){
		if (line.contains(startJoint))
			return startJoint;
		else if (line.contains(endJoint))
			return endJoint;
		return null;
	}
	public boolean contains(Joint joint){
		return (joint == startJoint || joint == endJoint);
	}
	public Vector3 getVector3(){
		return Vector3.subtract(end, start);
	}
	public Vector3 getNormal(){
		Vector3 normal = getVector3();
		normal.normalize();
		return normal;
	}
	

	

	
	
	public void addPlate(Plate plate){
		plates.addElement(plate);
	}	
	

	Line(Joint j0, Joint j1){
		startJoint = j0;
		endJoint = j1;
		
		start = j0.position;
		end = j1.position;
	
		
	}


	public boolean active;
	public void activate(){
		active = true;
		restoreStyle();
	}
	public void disactivate(){
		active = false;
		restoreStyle();
	}
	public void restoreStyle(){
		if (active)
			lineAppearance.setDiffuseColor(HIGHLIGHTED_COLOR);
		else
			lineAppearance.setDiffuseColor(BASE_COLOR);
	}
	
	
	
	
	
	public void delete(){

		lineVehicle.setParent(null);
		
		if (startJoint != null)
			startJoint.deleteLine(this);
		if (endJoint != null)
			endJoint.deleteLine(this);
		
	}

	public void set_candidate_style(){
		setCandidateGeometry();
		lineAppearance.setDiffuseColor(CANDIDATE_COLOR);//new Color(0.4,0.0,0.9));//0.7,0,1));
	}
	public void set_primary_candidate_style(){
		setCandidateGeometry();
		lineAppearance.setDiffuseColor(new Color(1,0,0));
	}
	
	public void set_regular_style(){
		setRegularGeometry();
		lineAppearance.setDiffuseColor(BASE_COLOR);//new Color(0.2,0.2,0.2));
	}



	public void setEndPosition(Vector3 p){
		end = p;

		Vector3 vec = getVector3();
		Vector3 up = getUp(vec);
		lineVehicle.setOrientation(vec, up, scene);
		lineVehicle.scale(new Vector3(1,1,vec.getLength()), lineVehicle); 
	}









	Transformable lineVehicle;
	Visual lineVisual;
	Appearance lineAppearance  = new Appearance();;

	public void removeFromScenegraph(){
		if (lineVehicle != null){
			lineVehicle.setParent(null);
		}
	}
	Scene scene;
	public void addToScenegraph(Scene scene){
		this.scene = scene;
		if (lineVehicle != null && lineVehicle.getParent() != null)
			return;
			
		if (lineVehicle == null){
			createLineVisual(start, end);
	
			lineVehicle = new Transformable();
			if (getVector3().getLength() != 0)
	        lineVisual.setParent(lineVehicle);

		    lineVehicle.setParent( scene );
			lineVehicle.setPosition(start, scene);
		
			Vector3 vec = getVector3();
			if (vec.getLength() == 0)
				vec = new Vector3(1,0,0);
			Vector3 up = getUp(vec);
			lineVehicle.setOrientation(vec, up, scene);
			lineVehicle.scale(new Vector3(1,1,vec.getLength()), lineVehicle); 

			set_regular_style();
		}
		else {
		    lineVehicle.setParent( scene );
			restoreStyle();
		}
	}
	public Transformable getVehicle(){
		return lineVehicle;
	}
	
		
	public void setJoints(Joint startJoint, Joint endJoint){
	
		this.startJoint = startJoint;
		this.endJoint = endJoint;
     
	}



	public void createLineVisual(Vector3 start, Vector3 end){
	
		lineVisual = new Visual();

//		lineVisual.setFillingStyle(FillingStyle.SOLID);
//		lineVisual.setShadingStyle(ShadingStyle.SMOOTH);//FLAT);
		lineAppearance = new Appearance();
		lineVisual.setFrontFacingAppearance(lineAppearance);
		lineAppearance.setFillingStyle(FillingStyle.SOLID);
		lineAppearance.setShadingStyle(ShadingStyle.SMOOTH);//FLAT);
		setRegularGeometry();
	}
	public void setRegularGeometry(){
		setGeometry();
	}
	public void setCandidateGeometry(){
		setGeometry_thin();
	}
	





	
	public Plane getNextPlane(Plane currentPlane){
	
		

		if (!currentPlane.contains(this)){
				Plane plane = newParallelPlane(currentPlane);
				if (plane == null){
					plane = getPlatePlane();
					if (plane == null)
						return null;
				}
				return plane;
		}

		Plane plane = newPerpendicularPlane(currentPlane);

		return plane;	



	}

	
	public Plane newPerpendicularPlane(Plane currentPlane){
		
		Vector3 new_base = start;
		Vector3 new_normal = Vector3.crossProduct(Vector3.subtract(end, start), currentPlane.normal);
		new_normal.normalize();
		
		return new Plane(new_base, new_normal);
	}

	public Plane newParallelPlane(Plane currentPlane){
		if (parallel(currentPlane.normal, getVector3()))
			return null;//newPlane();
			
		Vector3 perpendicularNormal = Vector3.crossProduct(getVector3(), currentPlane.normal);
		

		return new Plane(start, Vector3.crossProduct(getVector3(), perpendicularNormal));
	}
	public Plane getPlatePlane(){
		if (plates.size() ==0)
			return null;
		Plate plate = (Plate) plates.elementAt(0);
		return plate.getPlane();
		
	}


	// no base plane
	public Plane newPlane(){
		
		Vector3 vec = Vector3.subtract(end, start);
		Vector3 up = getUp(vec);
			
		Vector3 new_base = start;
		Vector3 new_normal = Vector3.crossProduct(vec, up);
		new_normal.normalize();
		
		Plane plane = new Plane(new_base, new_normal);
		//plane.addToScenegraph(Chateau.jaliceFrame.scene);
		
		return plane;
	}
	public Vector3 getUp(Vector3 vec){
		vec = Vector3.normalizeV(vec);
		if (Math.abs(Vector3.dotProduct(vec, Vector3.Y_AXIS)) < 0.99999)
			return Vector3.Y_AXIS;
		else
			return Vector3.X_AXIS;
		
	}









	public void setGeometry(){

		if (lineITA_normal == null)
			lineITA_normal = get_lineITA(0.05);
        lineVisual.setGeometry( lineITA_normal );

	}
	public void setGeometry_thin(){

		if (lineITA_thin == null)
			lineITA_thin = get_lineITA(0.04);
        lineVisual.setGeometry( lineITA_thin );
	}
	
	static IndexedTriangleArray lineITA_normal;
	static IndexedTriangleArray lineITA_thin;
	
	public IndexedTriangleArray get_lineITA(double radius){
		
		Vector3 start = new Vector3(0,0,0);
		Vector3 end = new Vector3(0,0,1);
		
		IndexedTriangleArray lineITA = new IndexedTriangleArray();

		Vector3[] v = new Vector3[4];
		
		Vector3 vec = Vector3.subtract(end, start);
		vec.normalize();
		Vector3 up = getUp(vec);

		Vector3 cross0 = Vector3.crossProduct(vec, up);
		cross0.normalize();
		Vector3 cross1 = Vector3.crossProduct(cross0, vec);
		
		Vector3[] top = new Vector3[8];
		Vector3[] bottom = new Vector3[8];
		
		for(int i=0; i<8; i++){
			double angle = Math.PI*2*i/8;
			Vector3 r = Vector3.add(Vector3.multiply(cross0, radius*Math.cos(angle)),
									Vector3.multiply(cross1, radius*Math.sin(angle)));
			top[i] = Vector3.add(start, r);	
			bottom[i] = Vector3.add(end, r);	
		}

		Vector3 nv;
		Vertex[] lineVertices = new Vertex[16+2];
		for(int i=0; i<8; i++){
			lineVertices[i] = new Vertex(Vertex.POSITION_XYZ|Vertex.NORMAL_IJK|Vertex.TEXTURE_COORDINATE_UV0);
			lineVertices[i].setXYZ(top[i]);
			nv = Vector3.subtract(top[i], start); nv.normalize();
			lineVertices[i].setIJK(nv);
			lineVertices[i].setU(0);
			lineVertices[i].setV(0);

			lineVertices[8+i] = new Vertex(Vertex.POSITION_XYZ|Vertex.NORMAL_IJK|Vertex.TEXTURE_COORDINATE_UV0);
			lineVertices[8+i].setXYZ(bottom[i]);
			nv = Vector3.subtract(bottom[i], end); nv.normalize();
			lineVertices[8+i].setIJK(nv);
			lineVertices[8+i].setU(0);
			lineVertices[8+i].setV(0);
		}
		lineVertices[16] = new Vertex(Vertex.POSITION_XYZ|Vertex.NORMAL_IJK|Vertex.TEXTURE_COORDINATE_UV0);
		lineVertices[16].setXYZ(start);
		lineVertices[16].setIJK(Vector3.multiply(vec, -1));
		lineVertices[16].setU(0);
		lineVertices[16].setV(0);

		lineVertices[17] = new Vertex(Vertex.POSITION_XYZ|Vertex.NORMAL_IJK|Vertex.TEXTURE_COORDINATE_UV0);
		lineVertices[17].setXYZ(end);
		lineVertices[17].setIJK(vec);
		lineVertices[17].setU(0);
		lineVertices[17].setV(0);
		
		
		
		
		int[] lineIndices = new int[3*8*2 + 3*8*2];

		for(int i=0; i<8; i++){
			lineIndices[i*6] = i;
			lineIndices[i*6+1] = 8+i;
			lineIndices[i*6+2] = Util.mod(i+1,8);
			
			lineIndices[i*6+3] = 8+i;
			lineIndices[i*6+4] = 8+Util.mod(i+1,8);
			lineIndices[i*6+5] = Util.mod(i+1,8);

			
			lineIndices[3*8*2+ i*6] = i;
			lineIndices[3*8*2+ i*6+1] = Util.mod(i+1,8);
			lineIndices[3*8*2+ i*6+2] = 16;
			
			lineIndices[3*8*2+ i*6+3] = 8+i;
			lineIndices[3*8*2+ i*6+4] = 17;
			lineIndices[3*8*2+ i*6+5] = 8+Util.mod(i+1,8);
			
		}
		
		

		lineITA.setVertices(lineVertices);
		lineITA.setIndices(lineIndices);
		
		return lineITA;
	
	}



}
