package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class MirrorHelper extends Helper
{
	public Suggestion makeSuggestion(Chateau chateau){
		
		if (chateau.currentPlane.is_ground)
			return null;
			
		if (chateau.currentLines.size() < 2)
			return null;
			
		Plane plane = chateau.currentPlane;
		
		Vector lines = chateau.currentLines;
		Vector pool = Util.duplicateVector(lines);

		if (all_on_plane(plane, lines))
			return null;
		
		Hashtable hash = new Hashtable();
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (pool.contains(line)){
				pool.removeElement(line);
				if (is_mirror(plane, line)){
					hash.put(line, line);
				}
				else {
					boolean succeed = false;
					for(int j=0; j<pool.size(); j++){
						Line line2 = (Line) pool.elementAt(j);
						if (is_mirror(plane, line, line2)){
							hash.put(line, line2);
							pool.removeElement(line2);
							succeed = true;
							break;
						}
					}
					if (succeed == false)
						return null;
				}
			}
		}

		int perfect = 0;
		for(Enumeration e = hash.keys(); e.hasMoreElements(); ){
			Line line0 = (Line) e.nextElement();
			Line line1 = (Line) hash.get(line0);
			Vector3 start = mirror(plane, line0.start);
			Vector3 end = mirror(plane, line0.end);
			if ((Joint.equivalent(start, line1.start) && Joint.equivalent(end, line1.end))
				||(Joint.equivalent(start, line1.end) && Joint.equivalent(end, line1.start)))
				perfect++;
		}
		if (perfect == hash.size()){
			System.out.println("Perfect Mirror");
			return null;
		}











		// let's make a suggestion!!	
		World originalWorld = 	new World(chateau);
		
		Vector newLines = new Vector();
		for(Enumeration e = hash.keys(); e.hasMoreElements(); ){
			Line line0 = (Line) e.nextElement();
			Line line1 = (Line) hash.get(line0);
			
			if (line0 == line1){
				if (!plane.contains(line0)){
					Joint joint0 = getFarJoint(plane, line0);
					Joint joint1 = chateau.findJoint(mirror(plane, joint0.position));

					if (line0.get_the_other_joint(joint0) != joint1){
						newLines.addElement(chateau.addLine(new Line(joint0, joint1)));
						chateau.deleteLine(line0);
					}
				}
			}
			else {
				Joint joint0f = getFarJoint(plane, line0);
				Joint joint0n = line0.get_the_other_joint(joint0f);
				
				Joint joint1f = getFarJoint(plane, line1);
				Joint joint1n = line1.get_the_other_joint(joint1f);
				
				if (plane.distance(joint0f.position) > plane.distance(joint1f.position)){
					Vector3 point1f = mirror(plane, joint0f.position);
					Vector3 point1n = mirror(plane, joint0n.position);
					
					if (!joint1f.equivalent(point1f) || !joint1n.equivalent(point1n)){
						joint1f = chateau.findJoint(point1f);
						joint1n = chateau.findJoint(point1n);
						newLines.addElement(chateau.addLine(new Line(joint1f, joint1n)));
						chateau.deleteLine(line1);
					}
				}
				else{
					Vector3 point0f = mirror(plane, joint1f.position);
					Vector3 point0n = mirror(plane, joint1n.position);
					
					if (!joint0f.equivalent(point0f) || !joint0n.equivalent(point0n)){
						joint0f = chateau.findJoint(point0f);
						joint0n = chateau.findJoint(point0n);
						newLines.addElement(chateau.addLine(new Line(joint0f, joint0n)));
						chateau.deleteLine(line0);
					}
				} 
			
			}
		}
		for(int i=0; i<newLines.size(); i++){
			Line line = (Line) newLines.elementAt(i);
			chateau.addCurrentLine(line);
		}





		Suggestion suggestion = new Suggestion(chateau, "Mirror");
		chateau.restoreWorld(originalWorld);
		return suggestion;
	}
	
	public boolean is_mirror(Plane plane, Line line){
		if (plane.contains(line)){
			return true;
		}
		
		if (is_mirror(plane, line.startJoint, line.endJoint)){
			return true;
		}
		
		return false;
	}
	public boolean is_mirror(Plane plane, Line line0, Line line1){
		Joint joint0f = getFarJoint(plane, line0);
		Joint joint0n = line0.get_the_other_joint(joint0f);
				
		Joint joint1f = getFarJoint(plane, line1);
		Joint joint1n = line1.get_the_other_joint(joint1f);
		
		return (is_mirror(plane, joint0f, joint1f) && is_mirror(plane, joint0n, joint1n));
	}
	public boolean is_mirror(Plane plane, Joint joint0, Joint joint1){
	
		if (joint0.equivalent(joint1)){
			if (plane.contains(joint0.position)){
				return true;
			}
			else
				return false;
		}
			
		Vector3 vec = Vector3.subtract(joint0.position, joint1.position);
		double angle = Util.angle180(vec, plane.normal);
		if (angle > 15 && angle < 180-15)	//if (!plane.perpendicular(vec))
			return false;
		
		double a = plane.signed_distance(joint0.position);
		double b = plane.signed_distance(joint1.position);
		
		double a_min = -b*0.7;
		double a_max = -b/0.7;
		if ((a-a_min)*(a-a_max)<0){
			return true;
		}
		
		return false;
	}
	public Joint getFarJoint(Plane plane, Line line){
		double start = plane.distance(line.start);
		double end = plane.distance(line.end);
		
		if (Math.abs(start-end)<Chateau.ERROR){
			start = Line.distance(plane.base, line.start);
			end = Line.distance(plane.base, line.end);
		}
		
		if (start>end)
			return line.startJoint;
		else
			return line.endJoint;
	}
	public Vector3 mirror(Plane plane, Vector3 p){
		Vector3 p_to_base = Vector3.subtract(plane.base, p);
		Vector3 p_to_foot = Vector3.multiply(plane.normal, Vector3.dotProduct(plane.normal, p_to_base));
		return Vector3.add(p, Vector3.multiply(p_to_foot, 2));
	
	}
	public boolean all_on_plane(Plane plane, Vector lines){
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (!plane.contains(line))
				return false;
		}
		return true;
	}			
	
	
}
