package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class MultiRepeatHelper extends PlaneHelper
{
	public Suggestion makeSuggestion(Chateau chateau){
		Vector clusters = get_clusters(chateau.currentLines);
		if (clusters.size() != 3)
			return null;
			
		Vector cluster0 = (Vector) clusters.elementAt(0);
		Vector cluster1 = (Vector) clusters.elementAt(1);
		Vector cluster2 = (Vector) clusters.elementAt(2);
		
		if (cluster0.size() != cluster1.size() || cluster0.size() != cluster2.size())
			return null;
			
		Vector3 center0 = get_center(cluster0);
		Vector3 center1 = get_center(cluster1);
		Vector3 center2 = get_center(cluster2);

	
		if (!Line.parallel(Vector3.subtract(center1, center0), Vector3.subtract(center2, center0)))
			return null;
		
		if (!match(cluster0, center0, cluster1, center1) || !match(cluster0, center0, cluster2, center2))
			return null;

		// bring cluster1 center	
		if (Vector3.dotProduct(Vector3.subtract(center1, center0), Vector3.subtract(center2, center0))<0){
			Vector cluster = cluster1;
			cluster1 = cluster0;
			cluster0 = cluster;
			Vector3 center = center1;
			center1 = center0;
			center0 = center;
		}
		else if (Vector3.dotProduct(Vector3.subtract(center1, center2), Vector3.subtract(center0, center2))<0){
			Vector cluster = cluster1;
			cluster1 = cluster2;
			cluster2 = cluster;
			Vector3 center = center2;
			center2 = center1;
			center1 = center;
		}

		// cluster2 should be distant than cluster0
		if (Vector3.subtract(center1, center0).getLength() > Vector3.subtract(center2, center1).getLength()){
			Vector cluster = cluster0;
			cluster0 = cluster2;
			cluster2 = cluster;
			Vector3 center = center0;
			center0 = center2;
			center2 = center;
		}

		int num = (int) (Vector3.subtract(center2, center0).getLength()/Vector3.subtract(center1, center0).getLength() +0.5);

		
		if (num<2)
			return null;
			
			
			
			
					
		








		// let us make a suggestion!!!
			
		World originalWorld = 	new World(chateau);

		/*
		Vector plates = new Vector();
		Vector plateLines = new Vector();
		for(int i=0; i<cluster1.size(); i++){
			Line line = (Line) cluster1.elementAt(i);
			for(int j=0; j<line.plates.size(); j++){
				Plate plate = (Plate) line.plates.elementAt(j);
				if (!plates.contains(plate) && Util.isSubsetVector(plate.lines, cluster1)){
					plates.addElement(plate);
					plateLines.addElement(Util.duplicateVector(plate.lines));
				}
			}
		}
		for(int j=0; j<plates.size(); j++){
			Plate plate = (Plate) plates.elementAt(j);
			chateau.deletePlate(plate);
		}
		*/
		// register plates in cluster0
		Vector plates = new Vector();
		Vector plateLines = new Vector();
		for(int i=0; i<cluster0.size(); i++){
			Line line = (Line) cluster0.elementAt(i);
			for(int j=0; j<line.plates.size(); j++){
				Plate plate = (Plate) line.plates.elementAt(j);
				if (!plates.contains(plate) && Util.isSubsetVector(plate.lines, cluster0)){
					plates.addElement(plate);
					plateLines.addElement(Util.duplicateVector(plate.lines));
				}
			}
		}
		
		// delete cluster1
		Vector plates1 = new Vector();
		for(int i=0; i<cluster1.size(); i++){
			Line line = (Line) cluster1.elementAt(i);
			for(int j=0; j<line.plates.size(); j++){
				Plate plate = (Plate) line.plates.elementAt(j);
				if (!plates1.contains(plate) && Util.isSubsetVector(plate.lines, cluster1)){
					plates1.addElement(plate);
				}
			}
		}
		for(int j=0; j<plates1.size(); j++){
			Plate plate = (Plate) plates1.elementAt(j);
			chateau.deletePlate(plate);
		}
		for(int j=0; j<cluster1.size(); j++){
			Line line = (Line) cluster1.elementAt(j);
			chateau.deleteLine(line);
		}



		
		Vector3 unit_vec = Vector3.multiply(Vector3.subtract(center2, center0), 1.0/num);
		for(int i=0; i<num-1; i++){
			
			Vector3 shift_vec = Vector3.multiply(unit_vec, i+1);
			
			Hashtable hash = new Hashtable();
			for(int j=0; j<cluster0.size(); j++){
				Line line = (Line) cluster0.elementAt(j);
				Vector3 start = Vector3.add(line.start, shift_vec);
				Vector3 end = Vector3.add(line.end, shift_vec);
				
				Joint startJoint = chateau.findJoint(start);
				Joint endJoint = chateau.findJoint(end);
				Line newLine = chateau.addLine(new Line(startJoint, endJoint));
				
				chateau.addCurrentLine(newLine);
				
				hash.put(line, newLine);
			}
			
			// plates
			
			for(int j=0; j<plateLines.size(); j++){
				Vector oldLines = (Vector) plateLines.elementAt(j);
				Vector newLines = new Vector();
				for(int k=0; k<oldLines.size(); k++){
					Line oldLine = (Line) oldLines.elementAt(k);
					Line newLine = (Line) hash.get(oldLine);
					newLines.addElement(newLine);
				}
				chateau.addPlate(new Plate(newLines));
			}
			
		}
		
		
		Suggestion suggestion = new Suggestion(chateau, "MultiRepeat");
		chateau.restoreWorld(originalWorld);
		return suggestion;
	}
	
	public Vector get_clusters(Vector lines){
		Vector clusters = new Vector();
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (!included(line, clusters)){
				Vector cluster = line.gatherConnectedLines(lines);
				clusters.addElement(cluster);
			}
		}
		return clusters;
	}
	public boolean included(Line line, Vector clusters){
		for(int i=0; i<clusters.size(); i++){
			Vector cluster = (Vector) clusters.elementAt(i);
			if (cluster.contains(line))
				return true;
		}
		return false;
	}
	public Vector3 get_center(Vector lines){
		Vector3 total = new Vector3();
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			total = Vector3.add(total, line.midPoint());
		}
		return Vector3.multiply(total, 1.0/lines.size());
	}
	public boolean match(Vector cluster0, Vector3 center0, Vector cluster1, Vector3 center1){
		cluster1 = Util.duplicateVector(cluster1);
		Vector3 dif = Vector3.subtract(center1, center0);
		for(int i=0; i<cluster0.size(); i++){
			Line line0 = (Line) cluster0.elementAt(i);
			Vector3 start = Vector3.add(line0.start, dif);
			Vector3 end = Vector3.add(line0.end, dif);
			Line line1 = find_matching_line(cluster1, start, end);
			if (line1 == null)
				return false;
			cluster1.removeElement(line1);
		}
		return true;
	}
	public Line find_matching_line(Vector lines, Vector3 start, Vector3 end){
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if ((Joint.equivalent(line.start, start) && Joint.equivalent(line.end, end)) ||
			    (Joint.equivalent(line.start, end) && Joint.equivalent(line.end, start)))
			    	return line;
		}
		return null;		
	}
	
}
