package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class ParallelIntervalsHelper extends PlaneHelper
{
	public Suggestion makeSuggestion(Chateau chateau){
		Vector lines = chateau.currentLines;
		if (lines.size() <4 || Util.mod(lines.size(),2)==1)
			return null;
			
		Plane plane = find_common_plane(lines);
		if (plane == null)
			return null;
			
		
		Vector remaining = Util.duplicateVector(lines);
		Vector line0s = new Vector();
		Vector line1s = new Vector();
		for(int i=0; i<lines.size(); i++){
			Line line0 = (Line) lines.elementAt(i);
			if (remaining.contains(line0)){
				remaining.removeElement(line0);
				Line line1 = get_closest_parallel(line0, remaining);
				if (line1 == null)
					return null;
				remaining.removeElement(line1);
				
				if (line0.getLength() < line1.getLength()){
					line0s.addElement(line0);
					line1s.addElement(line1);
				}
				else{
					line0s.addElement(line1);
					line1s.addElement(line0);
				}
			}
		}
		
		// check connectivity..
		for(int i=0; i<line0s.size(); i++){
			Line line0 = (Line) line0s.elementAt(i);
			for(int j=0; j<line1s.size(); j++){
				Line line1 = (Line) line1s.elementAt(j);
				if (line0.connected(line1))
					return null;
			}
				
		}
		
		
		
		double d=0;
		double[] ds = new double[line0s.size()];
		for(int i=0; i<line0s.size(); i++){
			Line line0 = (Line) line0s.elementAt(i);
			Line line1 = (Line) line1s.elementAt(i);
			ds[i] = line0.getDistance_as_infinite(line1.start);
			d += ds[i];
		}
		d /= line0s.size();
		
		if (!within_threthold(d, ds) || already_all_equal(d, ds))
			return null;
		
		// if a short line is a part of a plate, make it stable
		for(int i=0; i<line0s.size(); i++){
			Line line0 = (Line) line0s.elementAt(i);
			if (line0.plates.size() > 0){
				Vector swap = line1s;
				line1s = line0s;
				line0s = swap;
			}
		}
		
			
		


















		// let us make a suggestion!!!
			
		World originalWorld = 	new World(chateau);

		Hashtable infLines = new Hashtable();
		for(int i=0; i<line0s.size(); i++){
			Line line0 = (Line) line0s.elementAt(i);
			Line line1 = (Line) line1s.elementAt(i);
			infLines.put(line0, get_alignment(line0, line1, d));
		}


		Hashtable hash = new Hashtable();
		for(int i=0; i<line0s.size(); i++){
			Line line0 = (Line) line0s.elementAt(i);
			Line line1 = (Line) line1s.elementAt(i);
			
			Joint startJoint = (Joint) hash.get(line0.startJoint);
			if ( startJoint == null){
				Vector3 p = find_new_corner(line0.startJoint, line0, line1, line0s, infLines, d, ds[i]);
				startJoint = chateau.findJoint(p);
				hash.put(line0.startJoint, startJoint);
			}

			Joint endJoint = (Joint) hash.get(line0.endJoint);
			if (endJoint == null){
				Vector3 p = find_new_corner(line0.endJoint, line0, line1, line0s, infLines, d, ds[i]);
				endJoint = chateau.findJoint(p);
				hash.put(line0.endJoint, endJoint);
			}
			
			chateau.deleteLine(line0);
			
			Line newLine0 = chateau.addLine(new Line(startJoint, endJoint));
			chateau.addCurrentLine(newLine0);
		}

		
		Suggestion suggestion = new Suggestion(chateau, "Intervals");
		chateau.restoreWorld(originalWorld);
		return suggestion;
	}
	
	public Vector get_clusters(Vector lines){
		Vector clusters = new Vector();
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (!included(line, clusters)){
				Vector cluster = line.gatherConnectedLines(lines);
				clusters.addElement(cluster);
			}
		}
		return clusters;
	}
	public boolean included(Line line, Vector clusters){
		for(int i=0; i<clusters.size(); i++){
			Vector cluster = (Vector) clusters.elementAt(i);
			if (cluster.contains(line))
				return true;
		}
		return false;
	}
	
	
	public Line get_closest_parallel(Line base, Vector lines){
		Line result = null;
		double min = 1000000;
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (overlap_paralle(line, base)){	//line.parallel(base)){
				double d = base.getDistance_as_infinite(line.start);
				if (d<min){
					min = d;
					result = line;
				}
			}	
		}
		return result;
	
	}
	public boolean overlap_paralle(Line line0, Line line1){
		if (!line0.parallel(line1))
			return false;
			
		Vector3 vec = Vector3.subtract(line0.end, line0.start);
		vec.normalize();
		Vector3 start0_to_start1 = Vector3.subtract(line1.start, line0.start);
		Vector3 start0_to_end1 = Vector3.subtract(line1.end, line0.start);
		start0_to_start1.normalize();
		start0_to_end1.normalize();
		if (Vector3.dotProduct(vec, start0_to_start1)< Chateau.ERROR &&
			Vector3.dotProduct(vec, start0_to_end1)< Chateau.ERROR) 
				return false; 
		
		vec.multiply(-1);
		Vector3 end0_to_start1 = Vector3.subtract(line1.start, line0.end);
		Vector3 end0_to_end1 = Vector3.subtract(line1.end, line0.end);
		end0_to_start1.normalize();
		end0_to_end1.normalize();
		if (Vector3.dotProduct(vec, end0_to_start1)< Chateau.ERROR &&
			Vector3.dotProduct(vec, end0_to_end1)< Chateau.ERROR) 
				return false; 
		
		return true;


	}
	
	public InfiniteLine get_alignment(Line line, Line base, double d){
		double s = base.getDistance_as_infinite(line.start);
		Vector3 p = Vector3.add(base.start, Vector3.multiply(Vector3.subtract(line.start, base.start), d/s));
		return new InfiniteLine(p, base.getVector3());
	}
	public boolean already_all_equal(double d, double[] b){
		for(int i=0; i<b.length; i++){
			if (Math.abs(d-b[i])>Chateau.ERROR)
				return false;
		}
		return true;
	}
	public boolean within_threthold(double d, double[] ds){
		for(int i=0; i<ds.length; i++){
			if (ds[i]<d/1.5 || ds[i] >d*1.5)
				return false;
		}
		return true;
	}
	public Vector3 find_new_corner(Joint joint, Line line0, Line line1, Vector line0s, Hashtable infLines, double d, double ds){
		Line connectedLine = findConnectedLine(joint, line0, line0s);
		if (connectedLine == null){
			Vector3 foot = line1.getFoot(joint.position);
			return Vector3.add(foot, Vector3.multiply(Vector3.subtract(joint.position, foot), d/ds));
		}
		else {
			InfiniteLine infLine0 = (InfiniteLine) infLines.get(line0);
			InfiniteLine infLine1 = (InfiniteLine) infLines.get(connectedLine);
			return infLine0.crossPoint(infLine1);
			
		}
	}

	/*
	public Suggestion makeSuggestion(Chateau chateau){
		Vector lines = chateau.currentLines;
		if (lines.size()!=8)
			return null;
			
		Vector clusters = get_clusters(lines);
		if (clusters.size() != 2)
			return null;
			
		if (find_common_plane(lines) == null)
			return null;
			
		Vector cluster0 = (Vector) clusters.elementAt(0);
		Vector cluster1 = (Vector) clusters.elementAt(1);
		
		if (cluster0.size() != cluster1.size())
			return null;
		if (!is_loop(cluster0) || !is_loop(cluster1))
			return null; 

		Object[] objects = get_loop(cluster0);
		cluster0 = (Vector) objects[0];
		
		Vector _cluster1 = new Vector();
		for(int i=0; i<4; i++){
			Line line0 = (Line) cluster0.elementAt(i);
			Line line1 = get_closest_parallel(line0, cluster1);
			if (line1 == null || _cluster1.contains(line1))
				return null;
				
			_cluster1.addElement(line1);
		}	
		cluster1 = _cluster1;


		
		double d=0;
		double[] b = new double[4];
		for(int i=0; i<4; i++){
			Line line0 = (Line) cluster0.elementAt(i);
			Line line1 = (Line) cluster1.elementAt(i);
			b[i] = line0.getDistance_as_infinite(line1.start);
			d += b[i];
		}
		d /= 4;
		
		if (already_all_equal(d, b))
			return null;
			
			
		


















		// let us make a suggestion!!!
			
		World originalWorld = 	new World(chateau);

		// cluster0 is inside
		if (total_length(cluster0) > total_length(cluster0)){
			Vector cluster = cluster0;
			cluster0 = cluster1;
			cluster1 = cluster;
		}

		
		
		InfiniteLine[] infLines = new InfiniteLine[4];
		for(int i=0; i<4; i++){
			Line line0 = (Line) cluster0.elementAt(i);
			Line line1 = (Line) cluster1.elementAt(i);
			infLines[i] = get_alignment(line0, line1, d);
		}
		
		
		Vector3[] corners = new Vector3[4];
		for(int i=0; i<4; i++){
			corners[i] = infLines[i].crossPoint(infLines[Util.mod(i+1,4)]);
		}
		
		
		
		for(int i=0; i<4; i++){
			Line line1 = (Line) cluster1.elementAt(i);
			chateau.deleteLine(line1);
		}
		
		Joint[] joints = new Joint[4];
		for(int i=0; i<4; i++){
			joints[i] = chateau.findJoint(corners[i]);
		}
		
		for(int i=0; i<4; i++){
			Line newLine = chateau.addLine(new Line(joints[i], joints[Util.mod(i+1,4)]));
			chateau.addCurrentLine(newLine);
		}
		
		Suggestion suggestion = new Suggestion(chateau, "Frame");
		chateau.restoreWorld(originalWorld);
		return suggestion;
	}
	*/

}
