package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class Plane extends ChateauElement
{
	Vector3 base;
	Vector3 normal;

	public final Color GROUND_COLOR = Def.GROUND_COLOR;	//new Color(0.55, 0.93, 0.55);
	public final Color DRAWING_PLANE_COLOR = Def.DRAWING_PLANE_COLOR;	//new Color( 1,1,1,0.2 ) ;
	boolean is_ground = false;
	Vector lines = new Vector();
	
	Plane(){
	}
	
	public void setGround(){
		is_ground = true;
	}

	public void refresh_network(Hashtable hash){
		Vector revised_lines = new Vector();
		for(int i=0; i<lines.size(); i++){
			Line original_line = (Line) lines.elementAt(i);
			Line line = (Line) hash.get(original_line);

			revised_lines.addElement(line);
		}
		lines = revised_lines;
	}
	public Plane duplicate(){
		Plane plane = new Plane();
		
		plane.base = base;
		plane.normal = normal;
		plane.lines = lines;
		plane.is_ground = is_ground;
		
		plane.planeVehicle = planeVehicle;
		plane.gridVehicle = gridVehicle;
		plane.planeVisual = planeVisual;
		plane.gridVisual = gridVisual;
		
		return plane;
	}
	
	
	Plane(Vector3 base, Vector3 normal){
		this.base = base;
		this.normal = normal;
		normal.normalize();
		
	}
	
	
	
	public Line map(Line line){
		return new Line(map(line.start), map(line.end));
	}
	public Vector3 map(Vector3 p){
		Scene scene = Chateau.jaliceFrame.scene;
		
		if (planeVehicle == null){
			planeVehicle = new Transformable();
	        planeVehicle.setParent( scene );
    	    setPosition(scene);
			
			Vector3 mapped = map_sub(p, scene);
			
			planeVehicle.setParent(null);
			planeVehicle = null;
			return mapped;
		}
		else
			return map_sub(p, scene);
	}
	public Vector3 map_sub(Vector3 p, Scene scene){
		Transformable dammy = new Transformable();
		dammy.setParent(scene);
		dammy.setPosition(p,scene);
		Vector3 mapped = dammy.getPosition(planeVehicle);
		dammy.setParent(null);
		return mapped;
	}
	
	
	
	public boolean parallel(Plane plane){
		return (Math.abs(Vector3.crossProduct(normal, plane.normal).getLength()) < Chateau.ERROR);
	
	}
	public boolean parallel(Line line){
		return (Math.abs(Vector3.dotProduct(normal, line.getNormal())) < Chateau.ERROR);
	
	}
	public boolean perpendicular(Line line){
		return (Math.abs(Vector3.crossProduct(normal, line.getNormal()).getLength()) < Chateau.ERROR);
	
	}
	public boolean perpendicular(Vector3 vec){
		return (Math.abs(Vector3.crossProduct(normal, vec).getLength()/vec.getLength()) < Chateau.ERROR);
	
	}

	public boolean equivalent(Plane plane){
		if (Math.abs(Vector3.crossProduct(normal, plane.normal).getLength()) < Chateau.ERROR){
			Vector3 vec = Vector3.subtract(plane.base, base);
			if (Math.abs(Vector3.dotProduct(normal, vec)) < Chateau.ERROR)
				return true;
		}
		return false;
	
	}
	public boolean contains(Line line){
		return (contains(line.start) && contains(line.end));
	}
	public boolean contains(Vector3 v){
		Vector3 vec = Vector3.subtract(v, base);
		if (Math.abs(Vector3.dotProduct(normal, vec)) < Chateau.ERROR)
				return true;
		return false;
	}
	public boolean contains(Joint joint){
		return contains(joint.position);
	}
	public boolean contains(Plate plate){
		for(int i=0; i<plate.vertices.size(); i++){
			Vector3 v = (Vector3) plate.vertices.elementAt(i);
			if (!contains(v))
				return false;
		}
		return true;
	}
	public Vector3 get_cross_point(InfiniteLine line){
		// ((start + t * vec) - base) * normal = 0
		Vector3 vec = line.direction;
		Vector3 base_to_start = Vector3.subtract(line.base, base);

		double d = Vector3.dotProduct(vec, normal);
		if (Math.abs(d)<Chateau.ERROR)
			return null;
		double s = Vector3.dotProduct(base_to_start, normal);
		double t = -s/d;
		return Vector3.add(line.base, Vector3.multiply(vec, t));
	}
	public Vector3 get_cross_point(Line line){
		// ((start + t * vec) - base) * normal = 0
		Vector3 vec = line.getVector3();
		Vector3 base_to_start = Vector3.subtract(line.startJoint.position, base);

		double d = Vector3.dotProduct(vec, normal);
		//System.out.println("d "+d);
		if (Math.abs(d)<Chateau.ERROR)
			return null;
		double s = Vector3.dotProduct(base_to_start, normal);
		double t = -s/d;
		//System.out.println("t "+t);
		if (t<-Chateau.ERROR || t>1+Chateau.ERROR)
			return null;
		return Vector3.add(line.start, Vector3.multiply(vec, t));
	}
	public double signed_distance(Vector3 p){
		return Vector3.dotProduct(normal, Vector3.subtract(p, base));
	}
	public double distance(Vector3 p){
		return Math.abs(signed_distance(p));
	}
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
		

	Transformable planeVehicle;
	Visual planeVisual;

	public void hide(){
		if (planeVehicle != null){
			if (!is_ground)
				planeVisual.setParent(null);
			//gridVehicle.setParent(null);
		}
	}
	public void show(){

		if (planeVehicle != null){
			planeVisual.setParent(planeVehicle);
			//gridVehicle.setParent(planeVehicle);
		}
	}
	public void delete(){
		hide();
		if (planeVehicle != null){
			planeVehicle.setParent(null);
		}
	}

	public void addToScenegraph(Scene parent){
		if (planeVehicle == null){
			createPlaneVisual();

			planeVehicle = new Transformable();
   		    planeVehicle.setParent( parent );
        
       		setPosition(parent);
       		
       		//if (is_ground)
       		//	addVisibleGround();
       	}
       	else
   		    planeVehicle.setParent( parent );
       		
	}
	public void removeFromScenegraph(){
		if (planeVehicle != null)
   		planeVehicle.setParent( null );
       		
	}
	public Transformable getVehicle(){
		return planeVehicle;
	}
		
	
	public void addToScenegraph(Transformable parent){
        planeVehicle.setParent( parent );
        
        setPosition(parent);
    }
	public void setPosition(ReferenceFrame parent){
        planeVehicle.setPosition(base, parent);
        
        Vector3 head = getHead();
        planeVehicle.setOrientation(head, normal,parent);
	}
	public Vector3 getHead(){
        Vector3 head0 = Vector3.crossProduct(normal, Vector3.Y_AXIS);
        Vector3 head1 = Vector3.crossProduct(normal, Vector3.X_AXIS);

        Vector3 head;
        if (head0.getLength()>head1.getLength())
        	head = head0;
        else
        	head = head1;
        	
        return head;
	
	}
	public void moveTo(Vector3 pos, ReferenceFrame parent){
		//pos = projectOntoPlane(pos);
		if (contains(pos))
	        planeVehicle.setPosition(pos, parent);
	}
	
	






/*
	public void addVisibleGround(){
		Visual groundVisual = new Visual();
		
		groundVisual.setDiffuseColor(GROUND_COLOR);//	new Color(0,0.8,0,1) );//new Color(0.8, 0.8, 0.8));//
			
		groundVisual.setAmbientColor(new Color(1,0,0));
		
		groundVisual.setFillingStyle(FillingStyle.SOLID);
		groundVisual.setShadingStyle(ShadingStyle.FLAT);//NONE);
	
		groundVisual.setGeometry( ground_planeITA );
		
		
		
		Transformable groundVehicle = new Transformable();
		groundVehicle.setParent(planeVehicle);
		groundVisual.setParent(groundVehicle);
		
		groundVehicle.setPosition(new Vector3(0,-0.02,0), planeVehicle);
		
		
	}
	*/


	Appearance planeAppearance;
	public void createPlaneVisual()
	{
		planeVisual = new Visual();
		planeAppearance = new Appearance();
		planeVisual.setFrontFacingAppearance(planeAppearance);

		if (is_ground){
			planeAppearance.setDiffuseColor(GROUND_COLOR );
		}
        else{
        	planeAppearance.setDiffuseColor(DRAWING_PLANE_COLOR);
    		planeAppearance.setOpacity(0.5);

        }
		planeAppearance.setFillingStyle(FillingStyle.SOLID);
		planeAppearance.setShadingStyle(ShadingStyle.FLAT);//NONE);//
	
        if (is_ground){
        	if (ground_planeITA == null)
        		ground_planeITA = create_planeITA(4);
	        planeVisual.setGeometry( ground_planeITA );
        }
        else	{
        	if (planeITA == null)
        		planeITA = create_planeITA(8);
	        planeVisual.setGeometry( planeITA );
		}



		//createGrid();
	}
	
	
	int[][] plane_data = {{-1,0,-1},{-1,0,1},{1,0,-1}, {-1,0,1},{1,0,1},{1,0,-1}};


	static IndexedTriangleArray ground_planeITA;
	static IndexedTriangleArray planeITA;
	public IndexedTriangleArray create_planeITA(int size){
			
        IndexedTriangleArray planeITA = new IndexedTriangleArray();
		Vertex[] planeVertices = new Vertex[12];
		for(int i=0; i<6; i++){
			planeVertices[i] = new Vertex(Vertex.POSITION_XYZ|Vertex.NORMAL_IJK|Vertex.TEXTURE_COORDINATE_UV0);
			double vx = plane_data[i][0]*size;
			double vy = plane_data[i][1]*size;
			double vz = plane_data[i][2]*size;
			planeVertices[i].setXYZ(new Vector3(vx,vy,vz));
			planeVertices[i].setIJK(new Vector3(0,1,0));
			planeVertices[i].setU(0);
			planeVertices[i].setV(0);

			planeVertices[i+6] = new Vertex(Vertex.POSITION_XYZ|Vertex.NORMAL_IJK|Vertex.TEXTURE_COORDINATE_UV0);
			planeVertices[i+6].setXYZ(new Vector3(vx,vy,vz));
			planeVertices[i+6].setIJK(new Vector3(0,-1,0));
			planeVertices[i+6].setU(0);
			planeVertices[i+6].setV(0);
		}
		int[] planeIndices = {0,1,2,3,4,5 , 6,8,7,9,11,10};
		
		planeITA.setVertices(planeVertices);
		planeITA.setIndices(planeIndices);


		return planeITA;

	}












	Transformable gridVehicle;
	Visual gridVisual;
	Appearance gridAppearance;
	int[][] grid_data = {{0,0,0},{0,0,1},{1,0,0}, {0,0,1},{1,0,1},{1,0,0}};
	/*
	public void createGrid(int size, Color gridColor){
        gridVehicle = new Transformable();
        
		for(int x=-size; x<size; x++)
		for(int y=-size; y<size; y++)
		if ((x+y)/2*2==x+y)
		createGrid_sub(x,y,gridVehicle, gridColor);
    }
    
    
	public void createGrid_sub(int x, int y, Transformable gridVehicle, Color gridColor){
	*/
	public void createGrid(){
        gridVehicle = new Transformable();


        gridVisual = new Visual();
        gridVisual.setParent( gridVehicle );
        
        gridAppearance = new Appearance();
        gridVisual.setFrontFacingAppearance(gridAppearance);

        if (is_ground)
        	gridAppearance.setDiffuseColor(new Color(0,0.5,0,1));//new Color(0.5, 0.5, 0.5));//
		else
			gridAppearance.setDiffuseColor(new Color(0.5,0.5,0.5,0.2));
		
        gridAppearance.setFillingStyle(FillingStyle.SOLID);
        gridAppearance.setShadingStyle(ShadingStyle.FLAT);



        if (is_ground){
        	if (ground_gridITA == null)
        		ground_gridITA = create_gridITA(4);
 	       gridVisual.setGeometry( ground_gridITA );
        }
        else	{
        	if (gridITA == null)
        		gridITA = create_gridITA(8);
 	       gridVisual.setGeometry( gridITA );
		}
	}
	
	
	static IndexedTriangleArray ground_gridITA;
	static IndexedTriangleArray gridITA;
	public IndexedTriangleArray create_gridITA(int size){
		
	
        IndexedTriangleArray gridITA = new IndexedTriangleArray();

		Vertex[] gridVertices = new Vertex[12 * size*2 * size*2 / 2];
		int[] gridIndices = new int[12 * size*2 * size*2 / 2];
		int[] gridIndicesSub = {0,1,2,3,4,5, 6,8,7,9,11,10};

		int index = 0;
		
		for(int x=-size; x<size; x++)
		for(int y=-size; y<size; y++)
		if ((x+y)/2*2==x+y) {
			int base_index = index;
		for(int j=0; j<2; j++)
		for(int i=0; i<6; i++){
			gridVertices[index] = new Vertex(Vertex.POSITION_XYZ|Vertex.NORMAL_IJK|Vertex.TEXTURE_COORDINATE_UV0);
			double vx = grid_data[i][0]+x;
			double vy = grid_data[i][1]+0.001-0.002*j;
			double vz = grid_data[i][2]+y;
			gridVertices[index].setXYZ(new Vector3(vx,vy,vz));
			gridVertices[index].setIJK(new Vector3(0,1,0));
			gridVertices[index].setU(0);
			gridVertices[index].setV(0);
			
			gridIndices[index] = base_index + gridIndicesSub[index-base_index];

			index++;
		}
		}
		

		gridITA.setVertices(gridVertices);
		gridITA.setIndices(gridIndices);
		
		return gridITA;

	}













	
	
	
	public Vector3 get_local_coords(Vector3 v){
		return Chateau.jaliceFrame.get_local_coords(v, planeVehicle);
	}
	public Vector3 get_scene_coords(Vector3 v){
		return Chateau.jaliceFrame.get_scene_coords(v, planeVehicle);
	}














	/*
	public Vector predict(Line currentLine){
	
		Vector segments = new Vector();
		int focus = -1;
		for(int i=0; i<lines.size(); i++){
			Line line = (Line)lines.elementAt(i);

			if (line == currentLine)
				focus = i;

			Vector3 start = get_local_coords(line.start);
			Vector3 end = get_local_coords(line.end);
			
			double[] seg = {start.x, start.z, end.x, end.z};
			segments.addElement(seg);
			
		}

		Vector candidate_segments = pegasus.ChateauBind.predict(segments, focus);
		
		Vector candidate_lines = new Vector();
		for(int i=0; i<candidate_segments.size(); i++){
			double[] segment = (double[]) candidate_segments.elementAt(i);
			Vector3 u0 = new Vector3(segment[0], 0, segment[1]);
			Vector3 u1 = new Vector3(segment[2], 0, segment[3]);
			//System.out.println(""+u0+","+u1);
			
			u0 = Chateau.jaliceFrame.get_scene_coords(u0, planeVehicle);
			u1 = Chateau.jaliceFrame.get_scene_coords(u1, planeVehicle);
			Line line = new Line(this, u0, u1);
			candidate_lines.addElement(line);
		}
		
		return candidate_lines;
		

	}
	*/
}
