package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class Plate extends ChateauElement
{
	public int index;

	Vector lines;
	Vector vertices;
	
	public void refresh_network(Hashtable hash){
		Vector revised_lines = new Vector();
		for(int i=0; i<lines.size(); i++){
			Line original_line = (Line) lines.elementAt(i);
			Line line = (Line) hash.get(original_line);
			revised_lines.addElement(line);
		}
		lines = revised_lines;
		
	}
	
	public Plate duplicate(){
		Plate plate = new Plate();
		plate.lines = lines;
		plate.vertices = vertices;
		plate.plateVehicle = plateVehicle;
		plate.plateVisual = plateVisual;
		
		return plate;
	}
	public boolean equivalent(Plate plate){
		return (Util.equivalentVector(lines, plate.lines));
	}
	
	

	
	
	// this version does not support Plates with a hole...
	
	Plate(){};
	Plate(Vector lines){
		this.lines = lines;
		setVertices();
	}
	Plate(Line line0, Line line1, Line line2, Line line3){
		lines = new Vector();
		lines.addElement(line0);
		lines.addElement(line1);
		lines.addElement(line2);
		lines.addElement(line3);
		setVertices();
	}
	Plate(Line line0, Line line1, Line line2){
		lines = new Vector();
		lines.addElement(line0);
		lines.addElement(line1);
		lines.addElement(line2);
		setVertices();
	}	
	
	public void setVertices(){
		vertices = new Vector();
		
		Line baseLine = (Line) lines.elementAt(0);
		Line prevLine = baseLine;
		Joint prevJoint = baseLine.endJoint;
		vertices.addElement(prevJoint.position);
		while(true){
			prevLine = findNextLine(prevJoint, prevLine, lines);
			if (prevLine == baseLine)
				break;
			prevJoint = prevLine.get_the_other_joint(prevJoint);
			vertices.addElement(prevJoint.position);
		}
	}
	
	public Line findNextLine(Joint joint, Line prevLine, Vector lines){
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (line != prevLine && line.contains(joint))
				return line;
		}
		return null;
	}
	
	public Line findNextLine(Joint joint, Line prevLine){
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (line != prevLine && line.contains(joint))
				return line;
		}
		return null;
	}
	
	public void delete(){
		for(int i=0; i<lines.size(); i++)
			((Line)lines.elementAt(i)).plates.removeElement(this);
		plateVehicle.setParent(null);
	}

	public Plane getPlane(){
		
		return new Plane(getBase(), getNormal());
	}
	public Line get_common_line(Plate plate){
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (plate.lines.contains(line))
				return line;
		}
		return null;
	}
	public Vector3 getBase(){
		return (Vector3) vertices.elementAt(0);
	}
	public Vector3 getNormal(){
		Vector3 v0 = (Vector3) vertices.elementAt(0);
		Vector3 v1 = (Vector3) vertices.elementAt(1);
		Vector3 v2 = (Vector3) vertices.elementAt(2);
		Vector3 a = Vector3.subtract(v1,v0);
		Vector3 b = Vector3.subtract(v2,v0);
		Vector3 normal = Vector3.crossProduct(a,b);
		normal.normalize();
		return normal;
	}
	// this does not work for concave plate
	public boolean on_plate(Line line){
		return (on_plate(line.start) && on_plate(line.end));
	}
	public boolean on_plate(Vector3 p){
		Vector3 vec = Vector3.subtract(p, getBase());
		Vector3 normal = getNormal();
		if (Math.abs(Vector3.dotProduct(vec, normal))>Chateau.ERROR)
			return false;
	
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (line.on_line_segment(p))
				return true;
		}
		
		Vector3 v0 = Vector3.subtract((Vector3) vertices.elementAt(vertices.size()-1), p);
		double angle = 0;
		for(int i=0; i<vertices.size(); i++){
			Vector3 v1 = Vector3.subtract((Vector3) vertices.elementAt(i), p);
			angle += Util.signedAngle180(v0, v1, normal);
			v0 = v1;
		}
		if (angle > -180 && angle < 180)
			return false;
			
		return true;
	}
	public Vector3 base(){
		return (Vector3) vertices.elementAt(0);
	}
	
	public Vector3 get_cross_point(InfiniteLine lay){

		// ((start + t * vec) - base) * normal = 0
		Vector3 base = getBase();
		Vector3 normal = getNormal();
		
		Vector3 vec = lay.direction;
		Vector3 base_to_laybase = Vector3.subtract(lay.base, base);

		double d = Vector3.dotProduct(vec, normal);
		if (Math.abs(d)<Chateau.ERROR)
			return null;

		double s = Vector3.dotProduct(base_to_laybase, normal);
		double t = -s/d;
		
		Vector3 cross = Vector3.add(lay.base, Vector3.multiply(vec, t));
		if (!on_plate(cross))
			return null;
			
		return cross;
		
	}
	
	






	Transformable projectableVehicle;
	public void setProjectableVehicle(Scene parent){
		Vector3 v0 = (Vector3) vertices.elementAt(0);
		Vector3 v1 = (Vector3) vertices.elementAt(1);
		Vector3 v2 = (Vector3) vertices.elementAt(2);
		Vector3 a = Vector3.subtract(v1,v0);
		Vector3 b = Vector3.subtract(v2,v0);
		Vector3 normal = Vector3.crossProduct(a,b);
		normal.normalize();
	
	
		projectableVehicle = new Transformable();
		projectableVehicle.setParent(parent);
        projectableVehicle.setPosition(v0, parent);
        
        Vector3 head = getHead(normal);
        projectableVehicle.setOrientation(head, normal,parent);
	}
	public Vector3 getHead(Vector3 normal){
        Vector3 head0 = Vector3.crossProduct(normal, Vector3.Y_AXIS);
        Vector3 head1 = Vector3.crossProduct(normal, Vector3.X_AXIS);
        Vector3 head;
        if (head0.getLength()>head1.getLength())    head = head0;
        else										head = head1;
        return head;
	}














	Transformable plateVehicle;
	Visual plateVisual;
	Appearance plateAppearance;

	public void removeFromScenegraph(){
		if (plateVehicle != null){
			plateVehicle.setParent(null);
		}
		if (projectableVehicle != null){
			projectableVehicle.setParent(null);
		}
	}
	public void addToScenegraph(Scene scene){
		if (plateVehicle != null && plateVehicle.getParent() != null)
			return;
			
		if (plateVehicle == null){
			createPlateVisual();
		
			plateVehicle = new Transformable();
	        plateVisual.setParent(plateVehicle);
		}
		plateVehicle.setParent( scene );
		
		setProjectableVehicle(scene);
	}
	public Transformable  getVehicle(){
		return plateVehicle;
	}
		

	public void createPlateVisual(){

		plateVisual = new Visual();
		plateAppearance = new Appearance();
		plateVisual.setFrontFacingAppearance(plateAppearance);
		plateAppearance.setFillingStyle(FillingStyle.SOLID);
		plateAppearance.setShadingStyle(ShadingStyle.FLAT);
		
		setGeometry();
	}


	public void setGeometry(){

		plateITA = get_plateITA();
        plateVisual.setGeometry( plateITA );
        plateAppearance.setDiffuseColor(new Color(0.8,0.8,0.8));        

	}

	static IndexedTriangleArray plateITA;
	
	public IndexedTriangleArray get_plateITA(){

		/*
		Vector3 v0 = (Vector3) vertices.elementAt(0);
		Vector3 v1 = (Vector3) vertices.elementAt(1);
		Vector3 v2 = (Vector3) vertices.elementAt(2);
		Vector3 a = Vector3.subtract(v1,v0);
		Vector3 b = Vector3.subtract(v2,v0);
		Vector3 nv0 = Vector3.crossProduct(a,b);
		nv0.normalize();
		Vector3 nv1 = Vector3.multiply(nv0, -1);
		*/
		
		Vector3 v0 = (Vector3) vertices.elementAt(0);
		Vector3 nv0 = new Vector3();
		for(int i=0; i<vertices.size()-2; i++){
			Vector3 v1 = (Vector3) vertices.elementAt(i+1);
			Vector3 v2 = (Vector3) vertices.elementAt(i+2);
			Vector3 a = Vector3.subtract(v1,v0);
			Vector3 b = Vector3.subtract(v2,v0);
			Vector3 v = Vector3.crossProduct(a,b);
			nv0 = Vector3.add(nv0, v);
		}
		nv0.normalize();
		Vector3 nv1 = Vector3.multiply(nv0, -1);
		

		IndexedTriangleArray plateITA = new IndexedTriangleArray();

		Vertex[] plateVertices = new Vertex[vertices.size()*2];
		for(int i=0; i<vertices.size(); i++){
			Vector3 v = (Vector3) vertices.elementAt(i);

			plateVertices[i] = new Vertex(Vertex.POSITION_XYZ|Vertex.NORMAL_IJK|Vertex.TEXTURE_COORDINATE_UV0);
			plateVertices[i].setXYZ(v);
			plateVertices[i].setIJK(nv0);
			plateVertices[i].setU(0);
			plateVertices[i].setV(0);

			plateVertices[vertices.size()+i] = new Vertex(Vertex.POSITION_XYZ|Vertex.NORMAL_IJK|Vertex.TEXTURE_COORDINATE_UV0);
			plateVertices[vertices.size()+i].setXYZ(v);
			plateVertices[vertices.size()+i].setIJK(nv1);
			plateVertices[vertices.size()+i].setU(0);
			plateVertices[vertices.size()+i].setV(0);
		}

		int[][] triangles = delaunay(vertices, nv0);

		int[] plateIndices = new int[ triangles.length*3*2];
		for(int i=0; i<triangles.length; i++){
			int[] triangle = triangles[i];
			
			plateIndices[i*6] = triangle[0];
			plateIndices[i*6+1] = triangle[1];
			plateIndices[i*6+2] = triangle[2];

			plateIndices[i*6+3] = triangle[0] + vertices.size();
			plateIndices[i*6+4] = triangle[2] + vertices.size();
			plateIndices[i*6+5] = triangle[1] + vertices.size();
		}
		
		/*
		int[] plateIndices = new int[ (vertices.size()-2)*3*2];
		for(int i=0; i<vertices.size()-2; i++){
			plateIndices[i*6] = 0;
			plateIndices[i*6+1] = 1+i;
			plateIndices[i*6+2] = 2+i;

			plateIndices[i*6+3] = vertices.size();
			plateIndices[i*6+4] = vertices.size()+2+i;
			plateIndices[i*6+5] = vertices.size()+1+i;
		}
		*/
		
		
		plateITA.setVertices(plateVertices);
		plateITA.setIndices(plateIndices);
		
		return plateITA;
	}


	public int[][] delaunay(Vector vertices, Vector3 normal){
		Vector3 vec0 = getHead(normal);
		Vector3 vec1 = Vector3.crossProduct(normal, vec0);
		
		Vector loop = new Vector();
		for(int i=0; i<vertices.size(); i++){
			Vector3 v = (Vector3) vertices.elementAt(i);
			double x = Vector3.dotProduct(vec0, v);	
			double y = Vector3.dotProduct(vec1, v);	
			double[] xy = {x,y};
			loop.addElement(xy);
		}
		return teddy.Delaunay.triangulate_for_Plate(loop);
	}

}
