package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class Predictor {
	static public Vector predict(Vector currentLines){

		Vector predictedLines = new Vector();
		
		predictedLines = Util.connectVector(predictedLines, flip_connected_lines(currentLines));
		predictedLines = Util.connectVector(predictedLines, copy_context(currentLines));
		
		return predictedLines;
	}		
	static public Vector copy_context(Vector currentLines){
		Vector predictedLines = new Vector();

		if (currentLines.size() < 3)
			return predictedLines;		
		
		Line baseLine = (Line) currentLines.elementAt(currentLines.size()-1);	
		Vector3 baseVec = baseLine.getVector3();
		
		for(int i=0; i<currentLines.size()-1; i++){
			Line line = (Line) currentLines.elementAt(i);
			if (equivalent(baseVec, line.getVector3()))
				copy_context_sub(baseLine, line, currentLines, predictedLines);
		}

		return predictedLines;
	}
	static public boolean equivalent(Vector3 a, Vector3 b){
		if (!Line.parallel(a,b))
			return false;
		if (Math.abs(a.getLength()-b.getLength())>Chateau.ERROR)
			return false;
		return true;
	}
	static public void copy_context_sub(Line baseLine, Line referenceLine, Vector currentLines, Vector predictedLines){
		if (Vector3.dotProduct(baseLine.getVector3(),referenceLine.getVector3())>0){
			copy_context_sub_sub(referenceLine.startJoint, baseLine.startJoint, referenceLine, currentLines, predictedLines);
			copy_context_sub_sub(referenceLine.endJoint, baseLine.endJoint, referenceLine, currentLines, predictedLines);
		}
		else{
			copy_context_sub_sub(referenceLine.startJoint, baseLine.endJoint, referenceLine, currentLines, predictedLines);
			copy_context_sub_sub(referenceLine.endJoint, baseLine.startJoint, referenceLine, currentLines, predictedLines);
		}
	}
	static public void copy_context_sub_sub(Joint referenceJoint, Joint baseJoint, Line referenceLine, Vector currentLines, Vector predictedLines){
		for(int i=0; i<referenceJoint.lines.size(); i++){
			Line contextLine = (Line) referenceJoint.lines.elementAt(i);
			if (contextLine != referenceLine && currentLines.contains(contextLine)){
				Vector3 vector = Vector3.subtract(contextLine.get_the_other_joint(referenceJoint).position, 
													referenceJoint.position);
				Line predictedLine = new Line(baseJoint.position, Vector3.add(baseJoint.position, vector));
				predictedLines.addElement(predictedLine);
			}
		}	
	}
	
	
	
	static public Vector flip_connected_lines(Vector currentLines){
		Vector predictedLines = new Vector();

		if (currentLines.size() < 2)
			return predictedLines;		
			
		Line targetLine = (Line) currentLines.elementAt(currentLines.size()-2);
		Line baseLine = (Line) currentLines.elementAt(currentLines.size()-1);	
		
		Joint commonJoint = baseLine.get_common_joint(targetLine);
		
		if (commonJoint == null)
			return predictedLines;
		if (targetLine.parallel(baseLine))
			return predictedLines;
			
		
		Joint targetJoint = targetLine.get_the_other_joint(commonJoint);
		Vector3 targetPoint = targetJoint.position;
		Vector3 commonPoint = commonJoint.position;
		
		Vector3 vec =	baseLine.getVector3();
		vec.normalize();
		
		Vector3 flipped0 = flip(targetPoint, baseLine.start, vec);
		predictedLines.addElement(new Line(commonPoint, flipped0)); 
		
		Vector3 basePoint = baseLine.get_the_other_joint(commonJoint).position;
		Vector3 midPoint = baseLine.midPoint();
		Vector3 normal = Vector3.crossProduct(targetLine.getVector3(), vec);
		Vector3 vec2 = Vector3.crossProduct(normal, vec);
		vec2.normalize();
		
		Vector3 flipped1 = flip(targetPoint, midPoint, vec2);
		predictedLines.addElement(new Line(basePoint, flipped1)); 

		Vector3 flipped2 = flip(flipped0, midPoint, vec2);
		predictedLines.addElement(new Line(basePoint, flipped2)); 
		

		return predictedLines;
	}
	static public Vector3 flip(Vector3 p, Vector3 base, Vector3 vec){
		double t = Vector3.dotProduct(vec, Vector3.subtract(p, base));
		Vector3 foot = Vector3.add(base, Vector3.multiply(vec, t));
		return Vector3.add(foot, Vector3.subtract(foot, p));
	}


}