package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class PyramidHelper extends Helper
{
	public Suggestion makeSuggestion(Chateau chateau){
		if (chateau.currentLines.size() <= 3)
			return null;
	
		Line currentLine = chateau.currentLine();
		
		Joint commonJoint = null;
		Joint topJoint = null;
		if (currentLine.startJoint.lines.size() == 1 &&
			currentLine.endJoint.lines.size() >= 3) {
			commonJoint = currentLine.endJoint;
			topJoint = currentLine.startJoint;
		}
		else if (currentLine.startJoint.lines.size() >= 3 &&
			currentLine.endJoint.lines.size() == 1) {
			commonJoint = currentLine.startJoint;
			topJoint = currentLine.endJoint;
		}
		else
			return null;
		
		
		
		//Vector path = findPath(currentLine, commonJoint);		
		
		Vector baseLines = new Vector();
		for(int i=0; i<chateau.currentLines.size()-1; i++)
			baseLines.addElement(chateau.currentLines.elementAt(i));
		Plane basePlane = find_common_plane(baseLines);
		
		if (basePlane==null)
			return null;

		if (!is_loop(baseLines))
			return null;

		if (basePlane.contains(currentLine))
			return null;
		
		
		// let's make a suggestion!!	
		World originalWorld = 	new World(chateau);
		
		Joint newTopJoint;
		if (Line.parallel(currentLine.getVector3(), basePlane.normal)){
			Vector3 baseCenter = calculate_base_center(baseLines);
			Vector3 topPosition = Vector3.add(baseCenter, Vector3.subtract(topJoint.position, commonJoint.position));
			newTopJoint = chateau.findJoint(topPosition);

			chateau.deleteLine(currentLine);
		}
		else
			newTopJoint = topJoint;
		
		
		for(int i=0; i<baseLines.size(); i++){
			Line line = (Line) baseLines.elementAt(i);
		
			Line line0 = chateau.addLine(new Line(line.startJoint, newTopJoint));
			Line line1 = chateau.addLine(new Line(line.endJoint, newTopJoint));
		
			chateau.addCurrentLine(line0);	
			chateau.addCurrentLine(line1);	
			
			
			chateau.addPlate(new Plate(line, line0, line1));
			
		}
		
		
		//chateau.clearCurrentLines();
		chateau.switchPlane(chateau.ground);
		
		Suggestion suggestion = new Suggestion(chateau, "Pyramid");
		chateau.restoreWorld(originalWorld);
		return suggestion;		
	}


	public Vector3 calculate_base_center(Vector baseLines){
		Vector3 total = new Vector3();
		
		for(int i=0; i<baseLines.size(); i++){
			Line line = (Line) baseLines.elementAt(i);
			
			total.add(line.startJoint.position);
			total.add(line.endJoint.position);
		}
		total.multiply(1.0/(baseLines.size()*2));
		return total;
	}

}
