package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class RepeatHelper extends PlaneHelper
{
	public Suggestion makeSuggestion(Chateau chateau){
		int size = chateau.currentLines.size();
		if (size == 0 || size-(size/2)*2 != 0)
			return null;

		Line currentLine = chateau.currentLine();
		Line prevLine = (Line) chateau.currentLines.elementAt(chateau.currentLines.size()-2);

		Vector connected0 = prevLine.gatherConnectedLines(chateau.currentLines);
		Vector connected1 = currentLine.gatherConnectedLines(chateau.currentLines);
		
		if (connected0.size() != size/2 || connected1.size() != size/2)
			return null;
		


		// match currentLine and prevLine
		Vector3 vec0 = prevLine.getVector3();
		Vector3 vec1 = currentLine.getVector3();
		
		Vector3 base1;
		if (Vector3.add(vec0, vec1).getLength() < Chateau.ERROR) 
			base1 = currentLine.endJoint.position;
		else if (Vector3.subtract(vec0, vec1).getLength() < Chateau.ERROR) 
			base1 = currentLine.startJoint.position;
		else
			return null;
			
		Vector3 base0 = prevLine.startJoint.position;
		
		Vector3 shift_vec = Vector3.subtract(base1, base0);
		
		// match others
		for(int i=0; i<connected0.size(); i++){
			Line line0 = (Line) connected0.elementAt(i);
			Line line1 = find_corresponding_line(connected1, line0, shift_vec);
			if (line1 == null)
				return null;
		}
				
		




		// let us make a suggestion!!!
			
		World originalWorld = 	new World(chateau);

		chateau.clearCurrentLines();
		Line newCurrentLine = null;

		Vector plates = new Vector();
		Hashtable hash = new Hashtable();

		for(int i=0; i<connected1.size(); i++){
			Line line1 = (Line) connected1.elementAt(i);
			Vector3 newStart = Vector3.add(line1.startJoint.position, shift_vec);
			Vector3 newEnd = Vector3.add(line1.endJoint.position, shift_vec);
			
			Joint startJoint = chateau.findJoint(newStart);
			Joint endJoint = chateau.findJoint(newEnd);
			
			Line newLine = new Line(startJoint, endJoint);
			Line line2 = chateau.addLine(newLine);
			
			chateau.addCurrentLine(line1);
			chateau.addCurrentLine(line2);
			
			if (line1 == currentLine)
				newCurrentLine = line2;
				
				
			hash.put(line1, line2);
			for(int j=0; j<line1.plates.size(); j++){
				Plate plate = (Plate) line1.plates.elementAt(j);
				if (!plates.contains(plate))
					plates.addElement(plate);
			}			
		}		
		
		for(int i=0; i<plates.size(); i++){
			Plate plate = (Plate) plates.elementAt(i);
			Vector newLines = new Vector();
			for(int j=0; j<plate.lines.size(); j++){
				Line oldLine = (Line) plate.lines.elementAt(j);
				Line newLine = (Line) hash.get(oldLine);
				if (newLine == null){
					newLines = null;
					break;
				}
				newLines.addElement(newLine);
			}
			if (newLines != null)
				chateau.addPlate(new Plate(newLines));
		}
				
		
			
		chateau.addCurrentLine(currentLine);
		chateau.addCurrentLine(newCurrentLine);
		
		Suggestion suggestion = new Suggestion(chateau, "Repeat");
		chateau.restoreWorld(originalWorld);
		return suggestion;
	}
	
	
	public Line find_corresponding_line(Vector lines, Line base, Vector3 vec){
		Vector3 start = Vector3.add(base.startJoint.position, vec);
		Vector3 end = Vector3.add(base.endJoint.position, vec);
		
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (line.startJoint.equivalent(start) && line.endJoint.equivalent(end))
				return line;
			else if (line.startJoint.equivalent(end) && line.endJoint.equivalent(start))
				return line;
		}
		return null;
	}
}
