package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class RotationalRepeatHelper extends PlaneHelper
{
	public Suggestion makeSuggestion(Chateau chateau){
		Vector lines = chateau.currentLines;

		if (lines.size() != 3)
			return null;

		Plane basePlane = find_common_plane(lines);
		if (basePlane==null)
			return null;
		
		Object[] chain = get_chain(lines);
		if (chain == null)
			return radial(lines, basePlane, chateau);
		
		Line line0 = (Line)((Vector)chain[0]).elementAt(0);
		Line line1 = (Line)((Vector)chain[0]).elementAt(1);
		Line line2 = (Line)((Vector)chain[0]).elementAt(2);
		
		if (!line0.equalLength(line1) || !line0.equalLength(line2))
			return null;
			
		Joint joint0 = (Joint) ((Vector)chain[1]).elementAt(0);
		Joint joint1 = (Joint) ((Vector)chain[1]).elementAt(1);
		Joint joint2 = (Joint) ((Vector)chain[1]).elementAt(2);
		Joint joint3 = (Joint) ((Vector)chain[1]).elementAt(3);
		
		Vector3 vec0 = Vector3.subtract(joint1.position, joint0.position);
		Vector3 vec1 = Vector3.subtract(joint2.position, joint1.position);
		Vector3 vec2 = Vector3.subtract(joint3.position, joint2.position);
		
		double angle0 = Util.signedAngle180(vec0, vec1, basePlane.normal);
		double angle1 = Util.signedAngle180(vec1, vec2, basePlane.normal);
		
		if (Math.abs(angle0 - angle1) > Chateau.ERROR)
			return null;














		// let us make a suggestion!!!
			
		World originalWorld = 	new World(chateau);

		chateau.deleteLine(line1);
		chateau.deleteLine(line2);
		
		int n = (int)(360 / Math.abs(angle0) + 0.5);
		double angle = 360.0/n;
		if (angle0<0)
			angle *= -1;
		
		Joint[] joints = new Joint[n+1];
		joints[0] = joint0;
		joints[1] = joint1;
		joints[n] = joint0;
		for(int i=2; i<n; i++){
			Vector3 vec = Util.rotate(vec0, angle*(i-1), basePlane.normal);
			Vector3 p = Vector3.add(joints[i-1].position, vec);
			joints[i] = chateau.findJoint(p);
		}
		
		for(int i=1; i<n; i++){
			chateau.addCurrentLine(chateau.addLine(new Line(joints[i], joints[i+1])));
		}
		
		
		Suggestion suggestion = new Suggestion(chateau, "Rotation");
		chateau.restoreWorld(originalWorld);
		return suggestion;
	}
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	public Suggestion radial(Vector lines, Plane basePlane, Chateau chateau){
		Joint center = null;
		
		Line line0 = (Line)lines.elementAt(0);
		Line line1 = (Line)lines.elementAt(1);
		Line line2 = (Line)lines.elementAt(2);

		if (!line0.equalLength(line1) || !line0.equalLength(line2))
			return null;

		if (line1.contains(line0.startJoint) && line2.contains(line0.startJoint))
			center = line0.startJoint;
		else if (line1.contains(line0.endJoint) && line2.contains(line0.endJoint))
			center = line0.endJoint;
		else 
			return null;

		Vector3 vec0 = Vector3.subtract(line0.get_the_other_joint(center).position, center.position);
		Vector3 vec1 = Vector3.subtract(line1.get_the_other_joint(center).position, center.position);
		Vector3 vec2 = Vector3.subtract(line2.get_the_other_joint(center).position, center.position);
		
		double angle0 = Util.angle180(vec0, vec1);
		double angle1 = Util.angle180(vec1, vec2);
		double angle2 = Util.angle180(vec0, vec2);

		Line baseLine = null;
		double angle;
		if (Math.abs(angle0 - angle1) < Chateau.ERROR){
			baseLine = line0;
			angle = angle0;
		}
		else if (Math.abs(angle1 - angle2) < Chateau.ERROR){
			baseLine = line0;
			angle = angle1;
		}
		else if (Math.abs(angle0 - angle2) < Chateau.ERROR){
			baseLine = line1;
			angle = angle0;
		}
		else 
			return null;

















		// let us make a suggestion!!!
			
		World originalWorld = 	new World(chateau);
		
		if (line0 != baseLine)
			chateau.deleteLine(line0);
		if (line1 != baseLine)
			chateau.deleteLine(line1);
		if (line2 != baseLine)
			chateau.deleteLine(line2);
		
		
		
		int n = (int)(360 / angle + 0.5);
		angle = 360.0/n;
		
		Vector3 dvec = Vector3.subtract(baseLine.get_the_other_joint(center).position, center.position);
		
		for(int i=1; i<n; i++){
			Vector3 vec = Util.rotate(dvec, angle*i, basePlane.normal);
			Vector3 p = Vector3.add(center.position, vec);
			Joint joint = chateau.findJoint(p);
			chateau.addCurrentLine(chateau.addLine(new Line(center, joint)));
			
		}
		
		
		Suggestion suggestion = new Suggestion(chateau, "Rotation");
		chateau.restoreWorld(originalWorld);
		return suggestion;
	}
	
	
	
	
	
}
