package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class Snap {
	Vector joints;
	Vector points;
	Vector lines;
	Vector infiniteLines;
	Vector circles;
	Chateau chateau;
	Plane basePlane;
	
	Snap(Plane plane, Chateau chateau, double distance_to_pickpoint){
		this.chateau = chateau;
		this.basePlane = plane;
		this.snap_threshold = distance_to_pickpoint/40;// 10 -> 0.2
		
		joints = new Vector();
		points = new Vector();
		lines = new Vector();		
		circles = new Vector();		
		infiniteLines = new Vector();
		for(int i=0; i<chateau.joints.size(); i++){
			Joint joint = (Joint) chateau.joints.elementAt(i);
			if (plane.contains(joint.position)){
				joints.addElement(joint);
			}
		}
		for(int i=0; i<chateau.lines.size(); i++){
			Line line = (Line) chateau.lines.elementAt(i);
			if (plane.contains(line)){
				lines.addElement(line);
			}
			else {
				Vector3 p = plane.get_cross_point(line);
				if (p != null && !line.startJoint.equivalent(p) && !line.endJoint.equivalent(p) )
					points.addElement(p);
			}
		}
		
		
	}
	
	public void addInfiniteLine(Joint joint){
		for(int i=0; i<joint.lines.size(); i++){
			Line line = (Line) joint.lines.elementAt(i);
			if (basePlane.contains(line))
				addInfiniteLine_perpendicular(joint.position, line);
		}
	}
	public void addInfiniteLine_perpendicular(Vector3 base, Line line){
		Vector3 direction = Vector3.crossProduct(line.getVector3(), basePlane.normal);
		
		addInfiniteLine(new InfiniteLine(base, direction));
	}
	public void addInfiniteLine_parallel(Vector3 base, Line line){
		Vector3 direction = line.getVector3();
		addInfiniteLine(new InfiniteLine(base, direction));
	}
	public void addInfiniteLine(InfiniteLine infiniteLine){
		for(int i=0; i<infiniteLines.size(); i++){
			InfiniteLine line = (InfiniteLine) infiniteLines.elementAt(i);
			if (line.equivalent(infiniteLine))
				return;
		}	
		for(int i=0; i<infiniteLines.size(); i++){
			InfiniteLine line = (InfiniteLine) infiniteLines.elementAt(i);
			Vector3 cross = infiniteLine.crossPoint(line);
			if (cross != null)
				addPoint(cross);
		}
	
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			Vector3 cross = infiniteLine.crossPoint(line);
			if (cross != null)
				addPoint(cross);
		}
			
		for(int i=0; i<circles.size(); i++){
			Circle circle = (Circle) circles.elementAt(i);
			Vector3[] cross = circle.crossPoints(infiniteLine);
			if (cross != null){
				for(int j=0; j<cross.length; j++){
					addPoint(cross[j]);
				}
			}
		}

		infiniteLines.addElement(infiniteLine);
	}	
	public void addPoint(Vector3 p){
		for(int i=0; i<points.size(); i++){
			Vector3 point = (Vector3) points.elementAt(i);
			if (Joint.equivalent(p, point))
				return;
		}
		points.addElement(p);
	}

	public void addCircle(Vector3 center, Vector3 p){
		Circle circle = new Circle(center, Line.distance(center, p));

		for(int i=0; i<infiniteLines.size(); i++){
			InfiniteLine line = (InfiniteLine) infiniteLines.elementAt(i);
			Vector3[] cross = circle.crossPoints(line);
			if (cross != null){
				for(int j=0; j<cross.length; j++){
					addPoint(cross[j]);
				}
			}
		}
	
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			Vector3[] cross = circle.crossPoints(line);
			if (cross != null){
				for(int j=0; j<cross.length; j++){
					addPoint(cross[j]);
				}
			}
		}

		circles.addElement(circle);
	
	}







	
	public void prepare_guides_after_start(Vector3 p){
		// perpendicular to base line
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (line.on_line_segment(p))
				addInfiniteLine_perpendicular(p, line);
		}
		
		// duplicate, parallel
		for(int i=0; i<chateau.currentLines.size(); i++){
			Line line = (Line) chateau.currentLines.elementAt(i);
			if (basePlane.parallel(line)){
				Vector3 vec = line.getVector3();
				Vector3 point = Vector3.add(p, vec);
				addPoint(point);
				point = Vector3.subtract(p, vec);
				addPoint(point);
				
				addInfiniteLine_parallel(p, line);
				
			}
		}
		
		// radiual constraints 
		
		for(int i=0; i<chateau.currentLines.size(); i++){
			Line line = (Line) chateau.currentLines.elementAt(i);
			if (line.startJoint.equivalent(p))
				addCircle(line.start, line.end);
			else if (line.endJoint.equivalent(p))
				addCircle(line.end, line.start);
		}
		
		
	}







	public double snap_threshold = 0.2;
	public Vector3 snap(Vector3 p){
		Vector3 pos = p;
		
		double min = snap_threshold;
		Joint snappedJoint = null;
		for(int i=0; i<joints.size(); i++){
			Joint joint = (Joint) joints.elementAt(i);
			double d =Vector3.subtract(joint.position, p).getLength();
			if (d<min){
					snappedJoint = joint;
					min = d;
					pos = joint.position;
			}
		}
		if (min < snap_threshold){
			addInfiniteLine(snappedJoint);
			return pos;
		}



		for(int i=0; i<points.size(); i++){
			Vector3 point = (Vector3) points.elementAt(i);
			double d =Vector3.subtract(point, p).getLength();
			if (d<min){
					min = d;
					pos = point;
			}
		}
		if (min < snap_threshold)
			return pos;


		
		Line snappedLine = null;
		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			double d = line.getDistance(p);
			if (d<min){
				min = d;
				pos = line.getClosestPoint(p);
				
				if (Line.distance(line.midPoint(), p)<snap_threshold){
					pos = line.midPoint();
					snappedLine = line;
				}
			}
		}

		if (min < snap_threshold){
			if (snappedLine != null)
				addInfiniteLine_perpendicular(pos, snappedLine);
			return pos;
		}

		for(int i=0; i<infiniteLines.size(); i++){
			InfiniteLine line = (InfiniteLine) infiniteLines.elementAt(i);
			double d = line.getDistance(p);
			if (d<min){
				min = d;
				pos = line.getClosestPoint(p);
			}
		}

		for(int i=0; i<circles.size(); i++){
			Circle circle = (Circle) circles.elementAt(i);
			double d = circle.getDistance(p);
			if (d<min){
				min = d;
				pos = circle.getClosestPoint(p);
			}
		}
		
		return pos;
		
	}
	
	
	
	
	
	public boolean feedback_enabled = false;
	public void enable_feedback(){
		feedback_enabled = true;
	}
	public void paint(Graphics g){
		if (!feedback_enabled)
				return;
		
		for(int i=0; i<infiniteLines.size(); i++){
			InfiniteLine line = (InfiniteLine) infiniteLines.elementAt(i);
			Vector3 v0 = chateau.jaliceFrame.scene_coords_to_screen_coords(line.base);
			Vector3 v1 = chateau.jaliceFrame.scene_coords_to_screen_coords(Vector3.add(line.base, line.direction));
			Vector3 dv = Vector3.multiply(Vector3.subtract(v1, v0), 10);
			v1 = Vector3.add(v0, dv);
			Vector3 v2 = Vector3.subtract(v0, dv);
			
			Point p0 = new Point((int)v1.x, (int)v1.y);			
			Point p1 = new Point((int)v2.x, (int)v2.y);			
			
			g.setColor(java.awt.Color.red);
			g.drawLine(p0.x, p0.y, p1.x, p1.y);
		}
		
		for(int i=0; i<points.size(); i++){
			Vector3 point = (Vector3) points.elementAt(i);
			Vector3 v = chateau.jaliceFrame.scene_coords_to_screen_coords(point);
			
			Point p0 = new Point((int)v.x, (int)v.y);			
			
			g.setColor(java.awt.Color.red);
			g.drawOval(p0.x-5, p0.y-5, 10,10);
		}
		
		Vector3 base_to_camera = Vector3.subtract(basePlane.base, chateau.jaliceFrame.cameraVehicle.getPosition(null));
		Vector3 vx = Vector3.crossProduct(basePlane.normal, base_to_camera);
		vx.normalize();
		Vector3 vy = Vector3.crossProduct(basePlane.normal, vx);
		
		for(int i=0; i<circles.size(); i++){
			Circle circle = (Circle) circles.elementAt(i);
			Point p = null;
			for(int j=0; j<=circle_step; j++){
				double c = Math.cos(Math.PI*j*2/circle_step)*circle.radius;
				double s = Math.sin(Math.PI*j*2/circle_step)*circle.radius;
				Vector3 vxx = Vector3.multiply(vx, c);
				Vector3 vyy = Vector3.multiply(vy, s);
				Vector3 v = Vector3.add(circle.center, Vector3.add(vxx, vyy));
				v = chateau.jaliceFrame.scene_coords_to_screen_coords(v);
				Point q = new Point((int)v.x, (int)v.y);
				if (p!=null)
					g.drawLine(p.x, p.y, q.x, q.y);
				p = q;
				
			}
		}
		
	}
	private final int circle_step = 32;
}