package ui;

import edu.cmu.cs.stage3.alice.scenegraph.*;
import edu.cmu.cs.stage3.math.*;

import java.awt.Graphics;
import java.awt.Point;
import java.util.*;


public class StairsHelper extends Helper
{
	public Suggestion makeSuggestion(Chateau chateau){
		if (chateau.currentLines.size() <  6)
			return null;
	
		Object[] chain_set = get_chain(chateau.currentLines);
		
		if (chain_set == null)
			return null;
		

		Vector lines = (Vector) chain_set[0];
		Vector joints = (Vector) chain_set[1];

		Plane plane = find_common_plane(lines);
		if (plane == null)
			return null;


		//
		Line v_line0 = (Line) lines.elementAt(0);
		Line h_line0 = (Line) lines.elementAt(1);
		if (!v_line0.perpendicular(h_line0))
			return null;

		Joint j0 = (Joint) joints.elementAt(0);
		Joint j1 = (Joint) joints.elementAt(1);
		Joint j2 = (Joint) joints.elementAt(2);
		Vector3 v_vec0 = Vector3.subtract(j1.position, j0.position);
		Vector3 h_vec0 = Vector3.subtract(j2.position, j1.position);
		
		int n = (lines.size()+1)/2;
		double[] vs = new double[n];
		double[] hs = new double[lines.size()/2];
		vs[0] = v_vec0.getLength();
		hs[0] = h_vec0.getLength();
		for(int i=1; i<n; i++){
			Line v_line = (Line) lines.elementAt(i*2);
			if (!v_line0.parallel(v_line))
				return null;
			j0 = (Joint) joints.elementAt(i*2);
			j1 = (Joint) joints.elementAt(i*2+1);
			Vector3 v_vec = Vector3.subtract(j1.position, j0.position);
			if (Vector3.dotProduct(v_vec0, v_vec)<0)
				return null;
			vs[i] = v_vec.getLength();

			if (i*2+1 < lines.size()){
				Line h_line = (Line) lines.elementAt(i*2+1);
				if (!h_line0.parallel(h_line))
					return null;
				j2 = (Joint) joints.elementAt(i*2+2);
				Vector3 h_vec = Vector3.subtract(j2.position, j1.position);
				if (Vector3.dotProduct(h_vec0, h_vec)<0 )
					return null;
				hs[i] = h_vec.getLength();
			}
			
		}
		
		
		//
		double v_average = 0;
		for(int i=0; i<vs.length; i++)
			v_average += vs[i]/vs.length;
		
		double h_average = 0;
		for(int i=0; i<hs.length; i++)
			h_average += hs[i]/hs.length;


		
		// if already perfect stairs, reject!
		boolean perfect = true;
		for(int i=0; i<vs.length; i++)
			if (Math.abs(v_average - vs[i]) > Chateau.ERROR){
				perfect = false;
				break;
			}
		for(int i=0; i<hs.length; i++)
			if (Math.abs(h_average - hs[i]) > Chateau.ERROR){
				perfect = false;
				break;
			}
		if (perfect)
			return null;











		// let's make a suggestion!!	
		World originalWorld = 	new World(chateau);

		Vector3 new_v_vec = Vector3.multiply(v_vec0, v_average/vs[0]);
		Vector3 new_h_vec = Vector3.multiply(h_vec0, h_average/hs[0]);
		
		Joint[] newJoints = new Joint[joints.size()];
		newJoints[0] = (Joint) joints.elementAt(0);
		for(int i=0; i<n; i++){
			newJoints[i*2+1] = chateau.findJoint(Vector3.add(newJoints[i*2].position, new_v_vec));
			if (i*2+2 < joints.size())
				newJoints[i*2+2] = chateau.findJoint(Vector3.add(newJoints[i*2+1].position, new_h_vec));
		}

		chateau.clearCurrentLines();
		for(int i=0; i<n; i++){
			chateau.addCurrentLine(chateau.addLine(new Line(newJoints[i*2], newJoints[i*2+1])));
			if (i*2+1 < lines.size())
				chateau.addCurrentLine(chateau.addLine(new Line(newJoints[i*2+1], newJoints[i*2+2])));
		}


		for(int i=0; i<lines.size(); i++){
			Line line = (Line) lines.elementAt(i);
			if (!chateau.currentLines.contains(line))
				chateau.deleteLine(line);
		}



		Suggestion suggestion = new Suggestion(chateau, "Stairs");
		chateau.restoreWorld(originalWorld);
		return suggestion;
		
	}
}
